// calsrc.c - create THS file from PRB file

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <sigpro.h>
#include "mptrans.h"

#define MODNAM        "cavcal"
#define MAXLEN        80
#define MAXNCH        8
#define NSURGE        80
#define NFC           4
#define NTBP          5
#define NTAU          2
#define round(x)      ((int)floor((x)+0.5))
#define free_null(x)  if(x){free(x);x=NULL;}
#define max_(a,b)     ((a)>(b)?(a):(b))
#define min_(a,b)     ((a)<(b)?(a):(b))
#define fcopy(x,y,n)    memcpy(x,y,(n)*sizeof(float))
#define fzero(x,n)      memset(x,0,(n)*sizeof(float))
#define limit(a,b,c)    ((b<a)?a:((b>c)?c:b))

static double lmx = 12;     // maximum cavity length (cm)
static double lmn = 1;	    // minimum cavity length (cm)
static double mbw = 20;     // maximum band-width
static double fpr[2] = { 1, 10 };   // primary frequency range
static double fqm[2] = { 0.1, 10 }; // quality-metric frequency range
static double srg[2] = { 0, 0 };
static double ewc[2] = { 0, 0 };
static double dcv[2] = { 0, 0 };
static double lcv[2] = { 0, 0 };
static double zqm[2] = { 0, 0 };
static double ef1[2] = { 2000,  8000 };
static double ef2[2] = { 2000, 18000 };
static float *pcav, *zcav, *lcav;
static float *psrc = NULL;
static float *zsrc = NULL;
static float *pxtk = NULL;
static float* pc, * zc, * lc, * ps, * zs, * px, * pasv;
static float rate, attn, mpsn, pcer = 0, ewer = 0;
static float diam = 0.8f; // cavity diameter (cm)
static float temp = 25;   // cavity temperature (deg C)
static float zlk = 8000;  // cavity leak impedance (ohm)
static float rfl = 0.005f;// cavity flow loss (ohm)
static float tau[NTAU] = { 0 };       // ew time-coefficients
static float f_pri[2] = { 2, 10 };    // primary frequency range
static int wrdat = 0;     // write data ?
static int wrcav = 1;     // write cavity info ?
static int xtalk = 1;     // estimate cross-talk pressure ?
static int ewr = 1;       // perform evanescent-wave reduction ?
static int jef1, jef2, itrfl, itrlc;
static int ncav, nsrc, npts;

/***********************************************************************/


// write magnitude & phase (float array)
void
wrdbph(float *z, char *lab, int seq, int unw)
{
    char fn[20];
    double f, df;
    float *db, *ph;
    int  i, np;
    FILE *fp;

    df = rate / npts;
    np = npts / 2;
    db = (float *) calloc(np, sizeof(float));
    ph = (float *) calloc(np, sizeof(float));
    sp_cdb(z, db, np);
    sp_cph(z, ph, np);
    if (unw) {
        sp_unwrap(ph, ph, np);
    }
    sprintf(fn, "%s%02d.txt", lab, seq);
    fp = fopen(fn, "w");
    fprintf(fp, "; %s - from %s\n", fn, MODNAM);
    fprintf(fp, ";  f        db      ph\n");
    for (i = 0; i < np; i++) {
        f = i * df;
        fprintf(fp, "%6.0f %8.2f %8.3f\n", f, db[i], ph[i]);
    } 
    fclose(fp);
    free(db);
    free(ph);
}

// write magnitude & phase (double array)
void
wrdbphd(double *zd, char *lab, int seq, int unw)
{
    float *zf;
    int  i, np;

    np = npts / 2;
    zf = (float *) calloc(np, sizeof(float));
    for (i = 0; i < np; i++) {
        zf[i] = (float) zd[i];
    }
    wrdbph(zf, lab, seq, unw);
    free(zf);
}

// write cavity info to .txt file
void
write_cav(char *ifn, float *lc, double dba, double tgd, double hfr, double srd, 
	  double pc_err, double zc_err, int seq)
{
    char fn[20];
    int i;
    FILE *fp;

    // write lc
    sprintf(fn, "lc%02d.txt", seq);
    fp = fopen(fn, "w");
    fprintf(fp, "; %s - from %s\n", fn, MODNAM);
    for (i = 0; i < ncav; i++) {
        fprintf(fp, "%3.4f\n", lc[i]);
    }
    fclose(fp);
    // write dba, tgd, hfr
    sprintf(fn, "prbqm%02d.txt", seq);
    fp = fopen(fn, "w");
    fprintf(fp, "; %s - from %s\n", fn, MODNAM);
    fprintf(fp, "file=%s\n", ifn);
    fprintf(fp, "dBavg=%4.1f\n", dba);
    fprintf(fp, "delay=%4.2f\n", tgd);
    fprintf(fp, "dBhfr=%2.0f\n", hfr);
    fprintf(fp, "pc_err=%5.3f\n", pc_err);
    fprintf(fp, "zc_err=%5.3f\n", zc_err);
    fprintf(fp, "srgd=%6.3f\n", srd);
    fclose(fp);
}

// apply new extension to a filename
static void
newext(char *f, char *e)
{
    while (*f != '.' && *f != '\0')
	f++;
    *f++ = '.';
    while (*e != '\0')
	*f++ = *e++;
    *f++ = '\0';
}

/***********************************************************************/

// tube propagation - based on Allen (2020)
double
cav0(double f, double* Z, double* K)
{
    double  c, rho, mu, eta, beta, w, rc, dt, Zchr;
    static double eps = 1e-9;

    rc = diam / 2;      // radius of tube
    dt = temp - 26.85;  // temperature deviation
    w = 2 * M_PI * fmax(eps, f);
    c = 3.4723e4 * (1 + 0.00166 * dt);
    rho = 1.1769e-3 * (1 - 0.00335 * dt);
    // propagration function
    if (K) {
        mu = 18.6e-5;
        eta = 1.5345;
        beta = eta * sqrt(mu / rho) / rc;
        K[0] = (beta * sqrt(w / 2)) / c;
        K[1] = (w + beta * sqrt(w / 2)) / c;
    }
    // characteristic impedance
    Zchr = (rho * c) / (M_PI * rc * rc);
    if (Z) {
        Z[0] = Zchr;
        Z[1] = 0;
    }

    return (Zchr);
}

// ideal characterisic impedance
void
z_chr(float *z0)
{
    z0[0] = (float)cav0(0, 0, 0);
    z0[1] = 0;
}

double
get_term(float *lc)
{
    double rc, zo = 81;

    // this transformation improves pc_err ???
    rc = lc[ncav] / 1e6;
    rc = (rc < 0.999999) ? zo * (1 + rc) / (1 - rc) : 1e9;

    return (rc);
}

void
set_term(float *lc)
{
    // set initial value for terminal resistance
    lc[ncav] = itrfl ? 1e6F : 1e9F;
}

/***********************************************************************/

// vector sum
static double
vsum(double* a, int n)
{
    double sum = 0;
    int i;

    for (i = 0; i < n; i++) {
        sum += a[i];
    }
    return (sum);
}

// vector dot product
static double
vdot(double* a, double* b, int n)
{
    double sum = 0;
    int i;

    for (i = 0; i < n; i++) {
        sum += a[i] * b[i];
    }
    return (sum);
}

/***********************************************************************/

// frequency-domain window
static void
fd_window(double* z, int n, double fdw)
{
    double a, dp, p, w;
    int i, ir, ii;

    a = 0.16;       // Blackman window
    dp = (M_PI / n) / fdw;
    for (i = 1; i < n; i++) {
        ir = i * 2;
        ii = ir + 1;
        p = i * dp;
        if (p < M_PI) {
            w = (1 - a + cos(p) + a * cos(2 * p)) / 2;
        }
        else {
            w = 0;
        }
        z[ir] *= w;
        z[ii] *= w;
    }
}

// frequency-domain window of real
static void
fd_window_real(double* a, int n, double fdw)
{
    double* aa;
    int i;

    aa = calloc(n * 2, sizeof(double));
    for (i = 0; i < n; i++) {
        aa[i * 2] = a[i];
    }
    fd_window(aa, n, fdw);
    for (i = 0; i < n; i++) {
        a[i] = aa[i * 2];
    }
}

// frequency-domain reflectance
static void
reflect(double* Gr, double* Gi, double* Zr, double* Zi, double Z0, double* Ze, int n)
{
    double rnr, rni, rdr, rdi, rdm;
    int i;

    for (i = 0; i < n; i++) {
        rnr = Zr[i] - Z0;
        rdr = Zr[i] + Z0;
        if (Ze) {
            rni = rdi = Zi[i] - Ze[i];
        }
        else {
            rni = rdi = Zi[i];
        }
        rdm = rdr * rdr + rdi * rdi;
        Gr[i] = (rnr * rdr + rni * rdi) / rdm;
        Gi[i] = (rni * rdr - rnr * rdi) / rdm;
    }
}

// surge impedance
void
z_surge(double* Zl, int na, double* Zxc, double* Ze, double fdw)
{
    double* Gr, * Gi, * v0, * v1, * v5, * w, * w5, * Zr, * Zi;
    double  Z0, Z1, Z5, sv0, sv1, sv5;
    int     j, k, kr, ki;

    w = (double*)calloc(na, sizeof(double));
    w5 = (double*)calloc(na, sizeof(double));
    Zr = (double*)calloc(na, sizeof(double));
    Zi = (double*)calloc(na, sizeof(double));
    v0 = (double*)calloc(na, sizeof(double));
    v1 = (double*)calloc(na, sizeof(double));
    v5 = (double*)calloc(na, sizeof(double));
    Gr = (double*)calloc(na, sizeof(double));
    Gi = (double*)calloc(na, sizeof(double));
    for (k = 0; k < na; k++) {
        kr = 2 * k;
        ki = kr + 1;
        w[k] = (2 * M_PI / na) * k;
        w5[k] = pow(w[k], 5);
        v0[k] = 1;
        v1[k] = sin(w[k] / 2);
        v5[k] = sin(w[k] / 10);
        Zr[k] = Zl[kr];
        Zi[k] = Zl[ki];
    }
    fd_window_real(v0, na, fdw);
    fd_window_real(v1, na, fdw);
    fd_window_real(v5, na, fdw);
    sv0 = vsum(v0, na);
    sv1 = vsum(v1, na);
    sv5 = vsum(v5, na);
    for (k = 0; k < na; k++) {
        v0[k] /= sv0;
        v1[k] /= sv1;
        v5[k] /= sv5;
    }
    Z0 = vdot(Zr, v0, na);
    Z1 = fabs(vdot(Zi, w, na) / vdot(w, w, na)) / 3;
    Z5 = Z0 * 4e-4;
    for (k = 0; k < NSURGE; k++) {
        for (j = 0; j < na; j++) {
            Ze[j] = Z1 * w[j] + Z5 * w5[j];
        }
        reflect(Gr, Gi, Zr, Zi, Z0, Ze, na);
        Z0 *= (1 + vdot(Gr, v0, na));
        Z1 *= (1 + vdot(Gi, v1, na));
        Z5 *= (1 + vdot(Gi, v5, na));
    }
    for (k = 0; k < na; k++) {
        Ze[k] = Z1 * w[k] + Z5 * w5[k];
    }
    Zxc[0] = (float)Z0;
    Zxc[1] = (float)(Z1 / Z0);
    Zxc[2] = (float)pow(Z5 / Z0, 0.2);
    free(w);
    free(w5);
    free(Zr);
    free(Zi);
    free(v0);
    free(v1);
    free(v5);
    free(Gr);
    free(Gi);
}

/***********************************************************************/

// ideal characterisic impedance
double
c_sound()
{
    double c, dt;

    dt = temp - 26.85;  // temperature deviation
    c = 3.4723e4 * (1 + 0.00166 * dt);

    return (c);
}

/***********************************************************************/

// transform cavity pressure waveforms to spectra
int
cavity_pressure(VAR *v, int ic, int is, float *pc)
{
    char varnam[20];
    double sc;
    float scale, dacsn, attn, *st, *rs;
    int i, k;
    int32_t *r;
    int16_t *s, sets;

    k = ic * nsrc + is + 1;
    sprintf(varnam, "cal%02d", k);
    if (sp_var_find(v, varnam) < 0) {
	return (1);
    }
    st = (float *) calloc(npts + 2, sizeof(float));
    rs = (float *) calloc(npts + 2, sizeof(float));
    r = v[sp_var_find(v, varnam)].data;
    s = v[sp_var_find(v, "stim")].data;
    for (i = 0; i < npts; i++) {
        rs[i] = (float) r[i];
        st[i] = (float) s[i];
    }
    sp_transfer(st, rs, npts, pc);
    pc[0] = ic * 1e-9F;
    pc[1] = 0;
    scale = sp_var_f4(v, "scale");
    attn = sp_var_f4(v, "attn");
    strncpy(varnam, (is % 2) ? "da2sn" : "da1sn", 5);
    dacsn = (sp_var_find(v, varnam) < 0) ? 6922 : (float) sp_var_f4(v, varnam);
    sets = (sp_var_find(v, "sets") < 0) ? 2 : sp_var_i2(v, "sets");
    sc = scale * dacsn * pow(10, attn / 20) / sets;
    sp_smul(pc, pc, npts + 2, sc);
    free(rs);
    free(st);

    return (0);
}

// estimate cavity lengths from spectra
void
cavity_length(float *pc, float *lc)
{
    double c, d, t;
    float *p;
    int imn, imx, nf, m; 

    p = (float *) calloc(npts + 2, sizeof(float));
    nf = npts / 2 + 1;
    sp_cmagsq(pc, p, nf);
    sp_crfft(p, npts);
    c = c_sound();
    imn = round(lmn * rate * 2 / c);
    imx = round(lmx * rate * 2 / c);
    m = sp_vmax(p + imn, imx - imn) + imn;
    d = (p[m-1]-p[m+1])/(p[m-1]-2*p[m]+p[m+1])/2;
    t = (m + d) / rate;
    *lc = (float) (t  * c / 2);
    free(p);
}

// ideal-cavity impedance
static void
cavity_impedance(float* zc, int j, float lc, double rfl, float* tau)
{
    double e, c, s, nr, ni, dr, di, dm, Z[2], G[2];
    double cr, ci, sr, si, z0, f, w, ph, zpwr, zpwi, zfl, zew;
    int i;

    f = j ? j * (rate / npts) : 1e-9;
    w = 2 * M_PI * f;
    // planewave impedance
    cav0(f, Z, G);
    z0 = Z[0];
    e = exp(G[0] * lc);
    c = cos(G[1] * lc);
    s = sin(G[1] * lc);
    cr = c * e + c / e;
    ci = s * e - s / e;
    sr = c * e - c / e;
    si = s * e + s / e;
    nr = cr;
    ni = ci;
    dr = sr + cr * z0 / zlk;
    di = si + ci * z0 / zlk;
    dm = dr * dr + di * di;
    zpwr = z0 * (nr * dr + ni * di) / dm;
    zpwi = z0 * (ni * dr - nr * di) / dm;
    // flow loss
    zfl = -rfl * sqrt(f);
    // evanescent wave
    ph = 0;
    for (i = 0; i < NTAU; i++) {
        ph += pow(w * tau[i], 4 * i + 1);
    }
    zew = z0 * ph;
    // cavity impedance
    zc[0] = (float)(zpwr + zfl);
    zc[1] = (float)(zpwi + zew);
}

// cavity impedance for all cavities
void
compute_cavity_impedance(float* lc, float* tau, int j1, int j2)
{
    double rfl;
    int i, j, nf;

    nf = npts / 2 + 1;
    rfl = itrfl ? lc[ncav] : 0;
    for (j = j1; j < j2; j++) {
        for (i = 0; i < ncav; i++) {
            cavity_impedance(zc + (i * nf + j) * 2, j, lc[i], rfl, tau);
        }
    }
}

// write cavity data for plotting
void
plot_cavity(float *pc, float *zc, int sn)
{
    if (wrdat) {
        wrdbph(pc, "pc", sn, 1); 
        wrdbph(zc, "zc", sn, 0); 
    }
}

// compute dB average
double
db_avg(float *pc, int na, int nc)
{
    double p, sum, avg, eps = 1e-20;
    int i, n;

    n = na * nc;
    sum = 0;
    for (i = 0; i < n; i++) {
	p = fabs(pc[i]);
	p = (p < eps) ? eps : p;
	sum = sum + p * p;
    }
    avg = 2 * sum / n;
    return (10 * log10(avg));
}

// compute primary group delay (ms)
double
gd_fpr(float *pc, int na, float rate, double *prf)
{
    double pp, lp, gd, df, eps = 1e-20;
    double sx, sy, sxx, sxy, ff;
    int j, jj, jr, ji, n1, n2, nf;

    df = rate / (na - 2);
    n1 = (int) ((prf[0] / df) + 1 - eps);
    n2 = (int) ((prf[1] / df) + 1 - eps);
    nf = min_(n2, na) - n1;
    sx = sy = sxx = sxy = lp = 0;
    for (j = 0; j < nf; j++) {
	jj = n1 + j;
        jr = jj * 2;
        ji = jr + 1;
        pp = atan2(pc[ji], pc[jr]) / (2 * M_PI);
        while (pp < (lp - 0.5)) {
            pp = pp + 1;
        }
	while (pp > (lp + 0.5)) {
            pp = pp - 1;
        }
        lp = pp;
        ff = jj * df / 1000;
        sx += ff;
        sy += lp;
        sxx += ff * ff;
        sxy += ff * lp;
    }
    gd = -(nf * sxy - sx * sy) / (nf * sxx - sx * sx);
    return (gd);
}

// compute high-frequency dB range
double
db_hfr(float *pc, int na, int nc, float rate, float *mmf)
{
    double pm, mnp, mxp, df, eps = 1e-20;
    int i, j, jr, ji, k, n1, n2;

    df = (na - 2) / rate;
    n1 = round(mmf[0] * df);
    n2 = round(mmf[1] * df);
    n2 = min_(n2, na);
    mxp = eps;
    mnp = 1 / eps;
    for (i = 0; i < nc; i++) {
	k = i * na;
        for (j = n1; j < n2; j++) {
	    jr = k + j * 2;
	    ji = jr + 1;
            pm = pc[jr] * pc[jr] + pc[ji] * pc[ji];
	    if (mxp < pm) {
	        mxp = pm;
            }
            if (mnp > pm) {
	        mnp = pm;
            }
        }
    }
    if (mnp < eps) {
        mnp = eps;
    }
    return (10 * log10(mxp / mnp));
}

// compute reflectance group delay
double
reflectance_delay(float *Z, double z0, int nf, float rate, float *prf)
{
    double ph, pph, tpi, d, dsm, dss, f, fsm, fss, slp, gd;
    float df, rnr, rni, rdr, rdi, dms, Rr, Ri, Z0;
    int i, ir, ii, i1, i2, cnt;

    df = rate / (2 * (nf - 1));
    i1 = (int) max_(1, min_(nf - 1, floor(prf[0] / df)));
    i2 = (int) max_(1, min_(nf - 1, ceil(prf[1] / df)));
    tpi = 2 * M_PI;
    pph = dsm = dss = fsm = fss = 0;
    Z0 = (float) z0;
    for (i = i1 - 1; i < i2; i++) {
        ir = 2 * i;
        ii = ir + 1;
        rnr = Z[ir] - Z0;   
        rdr = Z[ir] + Z0;   
        rni = Z[ii];
        rdi = Z[ii];
        dms = rdr * rdr + rdi * rdi;
        Rr = (rnr * rdr + rni * rdi) / dms;
        Ri = (rni * rdr - rnr * rdi) / dms;
        ph = atan2(Ri, Rr) / tpi;
	while ((ph - pph) < -0.5) {
            ph += 1;
        } 
        while ((ph - pph) > 0.5) {
            ph -= 1;
        }
        pph = ph;
        if (i >= i1) {
	    f = i * df;
	    d = ph;
            fsm += f;
            fss += f * f;
            dsm += d;
            dss += d * f;
        }
    }
    cnt = i2 - i1;
    slp = (cnt * dss - dsm * fsm) / (cnt * fss - fsm * fsm);
    gd = - slp * 1000 ;

    return (gd);
}

// estimate Thevenin-equivalent source pressure & impedance 
void
thev_source(float *lc)
{
    double pir, pii, zir, zii, pms, zms, d;
    double a0, a1, a2, a3;
    double q0, q1, q2, q3;
    int i, ii, ir, k, kr, ki, nc, nf;
    static float eps = 1e-20F;

    nc = ncav;
    nf = npts / 2 + 1;
    // calculate ps & zs
    ps[0] = eps;
    zs[0] = eps;
    px[0] = eps;
    // loop over frequency
    for (k = 1; k < nf; k++) {
        // index real & imag
        kr = k * 2;
        ki = kr + 1;
        // zero summation variables
        a0 = a1 = a2 = a3 = 0;
        q0 = q1 = q2 = q3 = 0;
        // loop over cavity
        for (i = 0; i < nc; i++) {
            // index real & imag
            ir = (i * nf + k) * 2;
            ii = ir + 1;
            // get pc & zc for this frq, src, & cav
            pir = pc[ir];
            pii = pc[ii];
            zir = zc[ir];
            zii = zc[ii];
            // sum pc & zc combinations
            pms = pir * pir + pii * pii;
            zms = zir * zir + zii * zii;
            a0 += pms;
            a1 += zms;
            a2 += pir * zir + pii * zii;
            a3 += pii * zir - pir * zii;
            q0 += pir * zms;
            q1 += pii * zms;
            q2 -= zir * pms;
            q3 -= zii * pms;
        }
        // calculate determinant
        d = a1 * a0 - (a2 * a2 + a3 * a3);
        if (fabs(d) < eps) d = eps;
        // solve for ps & zs
        ps[kr] = (float) ((a0 * q0 + a2 * q2 - a3 * q3) / d);
        ps[ki] = (float) ((a0 * q1 + a2 * q3 + a3 * q2) / d);
        zs[kr] = (float) ((a1 * q2 + a2 * q0 + a3 * q1) / d);
        zs[ki] = (float) ((a1 * q3 + a2 * q1 - a3 * q0) / d);
    }
}

// write source data for plotting
void
plot_source(int sn)
{
    if (wrdat) {
        wrdbph(ps, "ps", sn, 1); 
        wrdbph(zs, "zs", sn, 0); 
        wrdbph(px, "px", sn, 1); 
    }
}

// compute Thevenin cavity-pressure error
double
pc_error(float* pa)
{
    double zzms, s1, s2;
    double zzr, zzi, pzr, pzi, ppr, ppi, pdr, pdi, par, pai;
    float psr, psi, zsr, zsi, pcr, pci, zcr, zci, * lc, * tau;
    int i, ir, ii, j, jr, ji, nf, nc, er;

    if (itrlc) {
        lc = pa;
        tau = pasv;
        for (i = 0; i < ncav; i++) if (lc[i] < 0.1) return (1e9);
        if (lc[i] < 0) return (1e9);
    }
    else {
        lc = pasv;
        tau = pa;
        for (i = 0; i < NTAU; i++) if (tau[i] < 0) return (1e9);
    }
    nc = ncav;
    nf = npts / 2 + 1;
    s1 = s2 = 0;
    compute_cavity_impedance(lc, tau, 0, nf);
    thev_source(lc);
    for (j = 0; j < nf; j++) {
        er = (j >= jef1 && j < jef2);
        jr = j * 2;
        ji = jr + 1;
        psr = ps[jr];
        psi = ps[ji];
        zsr = zs[jr];
        zsi = zs[ji];
        par = pai = 0;
        for (i = 0; i < nc; i++) {
            ir = (i * nf + j) * 2;
            ii = ir + 1;
            pcr = pc[ir];
            pci = pc[ii];
            zcr = zc[ir];
            zci = zc[ii];
            zzr = zsr + zcr;
            zzi = zsi + zci;
            pzr = psr * zcr - psi * zci;
            pzi = psi * zcr + psr * zci;
            zzms = zzr * zzr + zzi * zzi;
            ppr = (pzr * zzr + pzi * zzi) / zzms;
            ppi = (pzi * zzr - pzr * zzi) / zzms;
            pdr = pcr - ppr;
            pdi = pci - ppi;
            par = par + pdr;
            pai = pai + pdi;
            if (er) {
                s1 += pdr * pdr + pdi * pdi;
                s2 += pcr * pcr + pci * pci;
            }
        }
        if (xtalk) {
            px[jr] = (float)(par / nc);
            px[ji] = (float)(pai / nc);
            if (er) {
                s1 -= (par * par + pai * pai) / nc;
            }
        }
    }

    return ((s2 > 0) ? 1e4 * s1 / s2 : 0);
}

// print half-wave frequencies
void
prn_hw(float* lc, double tpc)
{
    double dt, c, f;
    int i;

    dt = temp - 26.85;  // temperature deviation
    c = 3.4723e4 * (1 + 0.00166 * dt) / 1000;
    printf("   fpk_hw:");
    for (i = 0; i < ncav; i++) {
        f = c / lc[i] / 2;
        printf(" %5.2f", f);
    }
    printf(" kHz; tpc=%.2f\n", tpc);
}

// print cavity lengths & Thevenin error
void
prn_pc(float* lc, double err)
{
    double rfl;
    int i;

    printf("   lc_fit:");
    for (i = 0; i < ncav; i++) {
        printf(" %5.2f", lc[i]);
    }
    rfl = itrfl ? lc[ncav] : 0;
    printf(" cm;  pc_err=%5.3f; rfl=%.3f\n", err, rfl);
}

// print cavity lengths & Thevenin error
void
prn_ew(float* tau, double err)
{
    int i;

    printf("   ew_fit:");
    for (i = 0; i < NTAU; i++) {
        printf(" %5.2f", tau[i] * 1e6);
    }
    printf(" us;  pc_err=%5.3f\n", err);
}

void
pc_fit(float* lc, double* erf, double df)
{
    double err;
    int nv, nf;
    OPT op;

    // set frequency range for pc_error calculaton
    jef1 = round(erf[0] / df);
    jef2 = round(erf[1] / df);
    nf = npts / 2 + 1;
    if (jef2 > nf) jef2 = nf;
    compute_cavity_impedance(lc, tau, jef1, jef2);
    thev_source(lc);
    // length iteration options
    op.icons = 0.000025F;
    op.ifrac = 0.01F;
    op.ffrac = 0.0001F;
    op.maxiter = 9000;
    op.miniter = ncav;
    op.escape = NULL;
    op.report = NULL;
    // ideal-cavity reflection
    nv = itrfl ? ncav + 1 : ncav;
    itrlc = 1;
    pasv = calloc(NTAU, sizeof(float));
    fcopy(pasv, tau, NTAU);
    sp_fmins(lc, nv, pc_error, &op);
    // estimate Thevenin-equivalent source pressure & impedance 
    compute_cavity_impedance(lc, tau, 0, nf);
    thev_source(lc);
    err = pc_error(lc);
    prn_pc(lc, err);
    free(pasv);
    pcer = (float)err;
}

void
ew_fit(float* lc, double* erf, double df)
{
    double err;
    int nf;
    OPT op;

    // set frequency range for pc_error calculaton
    jef1 = round(erf[0] / df);
    jef2 = round(erf[1] / df);
    nf = npts / 2 + 1;
    if (jef2 > nf) jef2 = nf;
    compute_cavity_impedance(lc, tau, jef1, jef2);
    thev_source(lc);
    // length iteration options
    op.icons = 0.000025F;
    op.ifrac = 0.01F;
    op.ffrac = 0.0001F;
    op.maxiter = 9000;
    op.miniter = ncav;
    op.escape = NULL;
    op.report = NULL;
    // ideal-cavity reflection
    itrlc = 0;
    pasv = calloc(ncav + 1, sizeof(float));
    fcopy(pasv, lc, ncav);
    sp_fmins(tau, NTAU, pc_error, &op);
    // estimate Thevenin-equivalent source pressure & impedance 
    compute_cavity_impedance(lc, tau, 0, nf);
    thev_source(lc);
    err = pc_error(tau);
    prn_ew(tau, err);
    free(pasv);
    ewer = (float)err;
}

// compute load impedance
static void
load_impedance(float *zl, int j, float *pk)
{
    double  plr, pli, pdr, pdi, pdm, prr, pri, zlr, zli;
    int     ii, ir;
    static float eps = 1e-9F;

    ir = 2 * j;
    ii = 2 * j + 1;
    plr = pk[ir];
    pli = pk[ii];
    pdr = ps[ir] - plr;
    pdi = ps[ii] - pli;
    pdm = pdr * pdr + pdi * pdi;
    prr = (plr * pdr + pli * pdi) / pdm;
    pri = (pli * pdr - plr * pdi) / pdm;
    zlr = zs[ir] * prr - zs[ii] * pri;
    zli = zs[ii] * prr + zs[ir] * pri;
    if (zlr < eps) {	// don't allow real(zl) <= 0
        zlr = eps;
    }
    zl[0] = (float) zlr;
    zl[1] = (float) zli;
}

// compute tube impedance
static void
tube_impedance(float *zt, int j, double lt, float *tp)
{
    double e, c, s, nr, ni, dr, di, dm, ctr, cti, Z[2], G[2];
    double cr, ci, sr, si, z0;
    double f, ff, f2, rc, z1r, z1i, z2r, z2i;

    // cavity impedance
    f = j * (rate / npts);
    rc = tp[0];
    cav0(f, Z, G);
    z0 = Z[0];
    e = exp(G[0] * lt);
    c = cos(G[1] * lt);
    s = sin(G[1] * lt);
    cr = c * e + c / e;
    ci = s * e - s / e;
    sr = c * e - c / e;
    si = s * e + s / e;
    nr = rc * cr + z0 * sr;
    ni = rc * ci + z0 * si;
    dr = z0 * cr + rc * sr;
    di = z0 * ci + rc * si;
    dm = dr * dr + di * di;
    ctr = (nr * dr + ni * di) / dm;
    cti = (ni * dr - nr * di) / dm;
    z1r = z0 * ctr;
    z1i = z0 * cti;
    // higer-order mode impedance
    ff = f / 16000;
    f2 = ff * ff;
    z2r = (-tp[1] + tp[3] * f2) * f2;
    z2i = (-tp[2] + tp[4] * f2) * f2 * ff;
    // sum higer-order mode impedance and tube impedance
    zt[0] = (float) (z1r + z2r);
    zt[1] = (float) (z1i + z2i);

    return;
}

// calculate load impedance for all frequencies & cavities
void
compute_load_impedance(float *zc, float *pc)
{
    int i, j, k, nf;

    nf = npts / 2 + 1;
    for (j = 0; j < nf; j++) {
        for (i = 0; i < ncav; i++) {
	    k = (i * nf + j) * 2;
            load_impedance(zc + k, j, pc);
	}
    }
}

// tube impedance for all cavities
void
compute_tube_impedance(float *zc, float *lc, float *tp)
{
    int i, j, nf;

    nf = npts / 2 + 1;
    for (j = 0; j < nf; j++) {
        for (i = 0; i < ncav; i++) {
            tube_impedance(zc + (i * nf + j) * 2, j, lc[i], tp);
	}
    }
}

// print quality metrics
void
print_qm(double dba, double tgd, double hfr)
{
    printf("dBavg=%4.1f; ", dba);
    printf("delay=%4.2f; ", tgd);
    printf("dBhfr=%2.0f\n", hfr);
}

// print cavity lengths & Thevenin error
void
print_length(char *lab, float *lc, double err)
{
    int i;

    printf("  %s lengths:", lab);
    for (i = 0; i < ncav; i++) {
        printf(" %5.2f", lc[i]);
    }
    printf(" cm;  err=%5.3f\n", err);
}

// print half-wave frequencies
void
print_hwfreq(float *lc)
{
    double dt, c, f;
    int i;

    dt = temp - 26.85;  // temperature deviation
    c = 3.4723e4 * (1 + 0.00166 * dt) / 1000;
    printf(" peak frequencies:");
    for (i = 0; i< ncav; i++) {
        f = c / lc[i] / 2;
        printf(" %5.2f", f);
    }
    printf(" kHz;  rc=%5.3f\n", get_term(lc));
}

// print cavity lengths & Thevenin error
void
print_hom(char *lab, float *tp, double err)
{
    int i;

    printf("    %s modes:", lab);
    printf(" %9.2e", tp[0]);
    for (i = 1; i < NTBP; i++) {
        printf(" %6.3f", tp[i]);
    }
    printf(";  zc_err=%5.3f\n", err);
}

/***********************************************************************/

static void
save_ths(float *zsrc, float *psrc, float *pxtk, float *lc,
    int nc, int ns, int np, int16_t sets,
    double rate, double df, double dc, double tc, double attn,
    const char *tfn)
{
    double *ps1, *ps2, *zs1, *zs2, *px1, *px2;
    float *lc1, *lc2;
    int i, j, nf, nv;
    int16_t na;
    VAR *vl;
    static char ftype[] = "SRC_THV";

    nf = np / 2 + 1;
    na = (int16_t) (np / 2);
    np = na * 2;
    // convert arrays from float to double
    ps1 = (double *) calloc(np, sizeof(double));
    ps2 = (double *) calloc(np, sizeof(double));
    zs1 = (double *) calloc(np, sizeof(double));
    zs2 = (double *) calloc(np, sizeof(double));
    px1 = (double *) calloc(np, sizeof(double));
    px2 = (double *) calloc(np, sizeof(double));
    lc1 = lc;
    lc2 = lc + nc;
    for (i = 0; i < np; i++) {
        j = i + nf * 2;
        ps1[i] = psrc[i];
        ps2[i] = psrc[j];
        zs1[i] = zsrc[i];
        zs2[i] = zsrc[j];
        px1[i] = pxtk[i];
        px2[i] = pxtk[j];
    }
    // set variables
    nv = 15;                // number of variables
    vl = sp_var_alloc(nv);
    sp_var_set(vl + 0,  ftype,  &na,   1, 1, "i2");
    sp_var_set(vl + 1,  "rate", &rate, 1, 1, "f8");
    sp_var_set(vl + 2,  "df",   &df,   1, 1, "f8");
    sp_var_set(vl + 3,  "dc",   &dc,   1, 1, "f8");
    sp_var_set(vl + 4,  "tc",   &tc,   1, 1, "f8");
    sp_var_set(vl + 5,  "zs1",  zs1,  na, 1, "f8c");
    sp_var_set(vl + 6,  "zs2",  zs2,  na, 1, "f8c");
    sp_var_set(vl + 7,  "ps1",  ps1,  na, 1, "f8c");
    sp_var_set(vl + 8,  "ps2",  ps2,  na, 1, "f8c");
    sp_var_set(vl + 9,  "px1",  px1,  na, 1, "f8c");
    sp_var_set(vl + 10, "px2",  px2,  na, 1, "f8c");
    sp_var_set(vl + 11, "lc1",  lc1,  nc, 1, "f4");
    sp_var_set(vl + 12, "lc2",  lc2,  nc, 1, "f4");
    sp_var_set(vl + 13, "attn", &attn, 1, 1, "f8");
    sp_var_set(vl + 14, "sets", &sets, 1, 1, "i2");
    // save file & clean-up
    sp_mat_save(tfn, vl);   // save variables in MAT file
    sp_var_clear(vl);
    free(ps1);
    free(ps2);
    free(zs1);
    free(zs2);
    free(px1);
    free(px2);
}

/***********************************************************************/

static void
zl_compute(double* zl, double* pl, double* pr, double* rl, double* zxc, double* ze,
    float* zs, float* ps, float* px, float* pc, int na, int sn)
{
    double  pfr, pfi, plr, pli, plm, prr, pri, psr, psi, por, poi;
    double  dnr, dni, dnm, zsr, zsi, z0, zlr, zli, fdw;
    double* Zsr, * Zsi, * Zlr, * Zli, * Gsr, * Gsi, * Gwr, * Gwi;
    int     i, ii, ir;
    static float eps = 1e-9F;

    Zsr = calloc(na, sizeof(double));
    Zsi = calloc(na, sizeof(double));
    Zlr = calloc(na, sizeof(double));
    Zli = calloc(na, sizeof(double));
    Gsr = calloc(na, sizeof(double));
    Gsi = calloc(na, sizeof(double));
    Gwr = calloc(na, sizeof(double));
    Gwi = calloc(na, sizeof(double));
    zl[0] = pl[0] = pr[0] = eps;
    zl[1] = pl[1] = pr[1] = 0;
    for (i = 1; i < na; i++) {
        ir = 2 * i;
        ii = 2 * i + 1;
        psr = ps[ir];
        psi = ps[ii];
        zsr = zs[ir];
        zsi = zs[ii];
        plr = pc[ir];
        pli = pc[ii];
        // load impedance: Zs*Pl/(Ps-Pl)
        dnr = psr - plr;
        dni = psi - pli;
        dnm = (dnr * dnr) + (dni * dni);
        prr = (plr * dnr + pli * dni) / dnm;
        pri = (pli * dnr - plr * dni) / dnm;
        zlr = zsr * prr - zsi * pri;
        zli = zsi * prr + zsr * pri;
        pl[ir] = plr;
        pl[ii] = pli;
        zl[ir] = zlr;
        zl[ii] = zli;
        // local copy
        Zsr[i] = zsr;
        Zsi[i] = zsi;
        Zlr[i] = zlr;
        Zli[i] = zli;
    }
    fdw = fmin(1, mbw * 2000 / rate);
    z_surge(zl, na, zxc, ze, fdw);
    z0 = zxc[0];
    reflect(Gsr, Gsi, Zsr, Zsi, z0, 0, na); // source reflectance
    reflect(Gwr, Gwi, Zlr, Zli, z0, ze, na); // waveguide reflectance
    for (i = 1; i < na; i++) {
        ir = 2 * i;
        ii = 2 * i + 1;
        // load pressure
        plr = pc[ir];
        pli = pc[ii];
        plm = plr * plr + pli * pli;
        // incident pressure: Po=Ps*(1-Gs)/2
        por = (ps[ir] * (1 - Gsr[i]) - ps[ii] * (0 - Gsi[i])) / 2;
        poi = (ps[ii] * (1 - Gsr[i]) + ps[ir] * (0 - Gsi[i])) / 2;
        // forward pressure: Pf=Po/(1-Gs.*Gw)
        dnr = 1 - (Gsr[i] * Gwr[i] - Gsi[i] * Gwi[i]);
        dni = 0 - (Gsi[i] * Gwr[i] + Gsr[i] * Gwi[i]);
        dnm = (dnr * dnr) + (dni * dni);
        pfr = (por * dnr + poi * dni) / dnm;
        pfi = (poi * dnr - por * dni) / dnm;
        // forward/load pressure ratio: Pf/Pl
        pr[ir] = (pfr * plr + pfi * pli) / plm;
        pr[ii] = (pfi * plr - pfr * pli) / plm;
        // waveguide impedance: Zw=Zo*(1+Gw)./(1-Gw)
        dnr = 1 - Gwr[i];
        dni = 0 - Gwi[i];
        dnm = (dnr * dnr) + (dni * dni);
        zl[ir] = z0 * ((1 + Gwr[i]) * dnr + (0 + Gwi[i]) * dni) / dnm;
        zl[ii] = z0 * ((0 + Gwi[i]) * dnr - (1 + Gwr[i]) * dni) / dnm;
        // waveguide reflectance: Gw
        rl[ir] = Gwr[i];
        rl[ii] = Gwi[i];
    }
    if (wrdat) {
        wrdbphd(zl, "zl", sn, 0);
        wrdbphd(pl, "pl", sn, 1);
        wrdbphd(pr, "pr", sn, 1);
    }
    free(Zsr);
    free(Zsi);
    free(Zlr);
    free(Zli);
    free(Gsr);
    free(Gsi);
    free(Gwr);
    free(Gwi);
}

// estimate cavity diameter from characteristic impedance
static double
cavity_diameter(double Zo)
{
    double c, rho, dt, A, D;

    dt = temp - 26.85;  // temperature deviation
    c = 3.4723e4 * (1 + 0.00166 * dt);
    rho = 1.1769e-3 * (1 - 0.00335 * dt);
    A = rho * c / Zo;
    D = 2 * sqrt(A / M_PI);

    return (D);
}

// estimate cavity length from reflectance 
static double
cavity_length_refl(double* G, int nf, double rate, double* fpr)
{
    double ph, dph, dly, f, df, len, ms;
    double sum1 = 0, sum2 = 0, lph = 0;
    int i, ir, ii;

    df = rate / (2 * (nf - 1)) / 1000;
    for (i = 0; i < nf; i++) {
        ir = 2 * i;
        ii = ir + 1;
        ph = atan2(G[ii], G[ir]) / (2 * M_PI);
        if (i > 0) { // unwrap phase
            while (ph > (lph + 0.5)) ph -= 1;
            while (ph < (lph - 0.5)) ph += 1;
        }
        dph = ph - lph;
        lph = ph;
        f = i * df;
        if ((f > fpr[0]) && (f < fpr[1])) {
            ms = G[ii] * G[ii] + G[ir] * G[ir];
            sum1 += dph * ms; // magnitude-squared weight
            sum2 += ms;
        }
    }
    dly = -(sum1 / sum2) / df;
    len = dly * c_sound() / 2000;
    return (len);
}

static double
zl_info(double* zl, double* pl, double* pr, double* rl, double* zxc, double* ze, int n, int sn)
{
    double Plr, Pli, Pfr, Pfi, Pwr, Pwi, Per, Pei, Pmr, Pmi, Pdr, Pdi;
    double dnr, dni, dnm, zrr, zri, sum1 = 0, sum2 = 0;
    int k, kr, ki, k1, k2;

    k1 = round(2000 * n * fqm[0] / rate);
    k2 = round(2000 * n * fqm[1] / rate);
    for (k = 0; k < n; k++) {
        kr = k * 2;
        ki = kr + 1;
        if ((k >= k1) && (k <= k2)) {
            // load pressure: Pl
            Plr = pl[kr];
            Pli = pl[ki];
            // forward pressure: Pf
            Pfr = Plr * pr[kr] - Pli * pr[ki];
            Pfi = Pli * pr[kr] + Plr * pr[ki];
            // waveguide pressure: Pw=Pf+Pr
            Pwr = Pfr + (Pfr * rl[kr] - Pfi * rl[ki]);
            Pwi = Pfi + (Pfi * rl[kr] + Pfr * rl[ki]);
            // evanescent pressure: Pe=Pw*(Ze/Zw)
            dnr = zl[kr];
            dni = zl[ki];
            dnm = (dnr * dnr) + (dni * dni);
            zrr = (ze[k] * dni) / dnm;
            zri = (ze[k] * dnr) / dnm;
            Per = Pwr * zrr - Pwi * zri;
            Pei = Pwi * zrr + Pwr * zri;
            // model pressure: Pm=Pw+Pe
            Pmr = Pwr + Per;
            Pmi = Pwi + Pei;
            // difference pressure: Pd=Pl-(Pw+Pe)
            Pdr = Plr - Pmr;
            Pdi = Pli - Pmi;
            // accumulate sums
            sum1 += (Pdr * Pdr) + (Pdi * Pdi);
            sum2 += (Plr * Plr) + (Pli * Pli);
        }
    }
    srg[sn - 1] = (float)zxc[0];
    ewc[sn - 1] = (float)zxc[1];
    dcv[sn - 1] = (float)cavity_diameter(zxc[0]);
    lcv[sn - 1] = (float)cavity_length_refl(rl, n, rate, fpr);
    zqm[sn - 1] = 1 - sum1 / sum2;
    return (zqm[sn - 1]);
}

static void
irdft(double* ft, int n, int nc, int ic, double fdw)
{
    double dp;
    int j, jn, k, kr, ki;

    dp = 2 * M_PI / n;
    for (j = 0; j < nc; j++) {
        jn = j * n * 2;
        for (k = 0; k < n; k++) {
            kr = k * 2;
            ki = kr + 1;
            ft[jn + kr] = cos(dp * ic * j) / n;
            ft[jn + ki] = sin(dp * ic * j) / n;
        }
        fd_window(ft + jn, n, fdw);
        ic++;
    }
}

static void
zl_repair(double* zl, double* zxc, int n)
{
    double a, b, df, fdw, zr, zi, zrmin, zimin, zmax, zm, ym, yr, yi;
    double* zrlim, * zilim, * ft;
    double zrsum = 0, zisum = 0, sum = 0;
    int j, jn, k, kr, ki, k1, k2, ic, nc;
    static double f1 = 80;  // very-low frequency
    static double f2 = 200; // low frequency

    zrlim = calloc(n, sizeof(double));
    zilim = calloc(n, sizeof(double));
    df = rate / (2 * n);
    // low frequency repair
    k1 = round(f1 / df);
    k2 = round(f2 / df);
    zmax = zxc[0] * 10;
    zrmin = zxc[0] / 100;
    zimin = -zmax;
    for (k = 0; k < k2; k++) {
        kr = k * 2;
        ki = kr + 1;
        zrlim[k] += limit(zrmin, zl[kr], zmax);
        zilim[k] += limit(zimin, zl[ki], zmax);
        zrsum += zrlim[k];
        zisum += zilim[k];
    }
    zr = zrsum / k2;
    zi = zisum / k2;
    for (k = 0; k < k1; k++) {
        kr = k * 2;
        ki = kr + 1;
        a = (1 + cos(M_PI * k / k1)) / 2;
        b = (1 - cos(M_PI * k / k1)) / 2;
        zl[kr] = a * zr + b * zrlim[k];
        zl[ki] = a * zi + b * zilim[k];
    }
    // flip negative-real admittance
    for (k = 0; k < k1; k++) {
        kr = k * 2;
        ki = kr + 1;
        zm = (zl[kr] * zl[kr]) + (zl[ki] * zl[ki]);
        yr = zl[kr] / zm;
        yi = -zl[ki] / zm;
        if (yr < 0) {
            ym = (yr * yr) + (yi * yi);
            zl[kr] = -yr / ym;
            zl[ki] = -yi / ym;
        }
    }
    // fix zero-frequency component
    ic = -16; // initial component
    nc = 9;   // number of components
    fdw = fmin(1, mbw * 2000 / rate);
    ft = calloc(n * nc * 2, sizeof(double));
    irdft(ft, n, nc, ic, fdw);
    for (j = 0; j < nc; j++) {
        jn = j * n * 2;
        sum = 0;
        for (k = 1; k < n; k++) {
            kr = k * 2;
            ki = kr + 1;
            sum += zl[kr] * ft[jn + kr];
            sum += zl[ki] * ft[jn + ki];
        }
    }
    free(ft);
    zl[0] = -sum / nc;
    zl[1] = 0;
}

static void
save_thl(float* zs1, float* ps1, float* px1, float* pc1, float* pc2,
    int na, int nc, int ns, int np, int16_t sets,
    double rate, double df, double at, const char* tfn)
{
    double* pl1, * pl2, * pr1, * pr2, * rl1, * rl2, * zl1, * zl2, * ze1, * ze2, * z0;
    double zxc1[3], zxc2[3], qm1, qm2;
    float* zs2, * ps2, * px2, zc[2];
    int i, ir, ii, n, nv;
    VAR* vl;
    static char ftype[] = "LD_THV";

    // convert arrays from float to double
    z0 = (double*)calloc(np, sizeof(double));
    zl1 = (double*)calloc(np, sizeof(double));
    zl2 = (double*)calloc(np, sizeof(double));
    ze1 = (double*)calloc(np, sizeof(double));
    ze2 = (double*)calloc(np, sizeof(double));
    pl1 = (double*)calloc(np, sizeof(double));
    pl2 = (double*)calloc(np, sizeof(double));
    pr1 = (double*)calloc(np, sizeof(double));
    pr2 = (double*)calloc(np, sizeof(double));
    rl1 = (double*)calloc(np, sizeof(double));
    rl2 = (double*)calloc(np, sizeof(double));
    z_chr(zc);
    n = np / 2;
    for (i = 0; i < n; i++) {
        ir = i * 2;
        ii = ir + 1;
        z0[ir] = zc[0];
        z0[ii] = zc[1];
    }
    zs2 = zs1 + na;
    ps2 = ps1 + na;
    px2 = px1 + na;
    zl_compute(zl1, pl1, pr1, rl1, zxc1, ze1, zs1, ps1, px1, pc1, n, 1);
    zl_compute(zl2, pl2, pr2, rl2, zxc2, ze2, zs2, ps2, px2, pc2, n, 2);
    qm1 = zl_info(zl1, pl1, pr1, rl1, zxc1, ze1, n, 1);
    qm2 = zl_info(zl2, pl2, pr2, rl2, zxc2, ze2, n, 2);
    zl_repair(zl1, zxc1, n);
    zl_repair(zl2, zxc2, n);
    // set variables
    nv = 14;                // number of variables
    vl = sp_var_alloc(nv);
    sp_var_set(vl + 0, ftype, &n, 1, 1, "i2");
    sp_var_set(vl + 1, "rate", &rate, 1, 1, "f8");
    sp_var_set(vl + 2, "df", &df, 1, 1, "f8");
    sp_var_set(vl + 3, "z0", z0, n, 1, "f8c");
    sp_var_set(vl + 4, "zl1", zl1, n, 1, "f8c");
    sp_var_set(vl + 5, "pl1", pl1, n, 1, "f8c");
    sp_var_set(vl + 6, "pr1", pr1, n, 1, "f8c");
    sp_var_set(vl + 7, "zl2", zl2, n, 1, "f8c");
    sp_var_set(vl + 8, "pl2", pl2, n, 1, "f8c");
    sp_var_set(vl + 9, "pr2", pr2, n, 1, "f8c");
    sp_var_set(vl + 10, "attn", &at, 1, 1, "f8");
    sp_var_set(vl + 11, "sets", &sets, 1, 1, "i2");
    sp_var_set(vl + 12, "qm1", &qm1, 1, 1, "f8");
    sp_var_set(vl + 13, "qm2", &qm2, 1, 1, "f8");
    // save file & clean-up
    sp_mat_save(tfn, vl);   // save variables in MAT file
    sp_var_clear(vl);
    free(z0);
    free(zl1);
    free(zl2);
    free(ze1);
    free(ze2);
    free(pl1);
    free(pl2);
    free(pr1);
    free(pr2);
}

/***********************************************************************/

static double
pc_delay(float* zs, double z0)
{
    double df, slp, pcd, tpi, f, zsr, zsi, zms, ysi, sx, sy, sxx, sxy;
    int i, ir, ii, i1, i2, cnt;

    df = rate / npts;   // frequency step
    i1 = round(f_pri[0] * 1000 / df);
    i2 = round(f_pri[1] * 1000 / df);
    cnt = 0;
    sx = sy = sxx = sxy = 0;
    for (i = i1; i < i2; i++) {
        ir = i * 2;
        ii = ir + 1;
        zsr = zs[ir];
        zsi = zs[ii];
        zms = zsr * zsr + zsi * zsi;
        ysi = -zsi / zms;
        f = i * df;
        sx += f;
        sy += ysi;
        sxx += f * f;
        sxy += f * ysi;
        cnt++;
    }
    tpi = 2 * M_PI;
    slp = (cnt * sxy - sx * sy) / (cnt * sxx - sx * sx);
    pcd = z0 * slp / tpi;
    return (pcd * 1000);
}

int
calsrc(char *ifn, double *erf, int thl, int wd, int stc, int hpf, int rr)
{
    char tfn[MAXLEN], *dot;
    double df, dba, hfr, tgd, z0, srd, pcd;
    float *pc1, *pc2, tp[NTBP];
    float  fco[1], A[NFC], B[NFC], *f, *h;
    int i, j, k, nf, na;
    int16_t sets;
    VAR *v;
    static char ftype[] = "EMAV_CAL";
    static float mmf[] = { 15000, 18000 };
    static float zcf[] = { 8000, 19000};
    static float prf[] = { 2000, 10000};

    // open PRB file
    v = sp_mat_load(ifn);
    if (!v) {
        fprintf(stderr, "ERROR: can't open %s.\n", ifn);
        return (1);
    }
    if (strcmp(v[0].name, ftype)) {
        fprintf(stderr, "ERROR: %s file type is not %s.\n", ifn, ftype);
        return (1);
    }
    ncav = sp_var_i2(v, "ncav");
    if (ncav < 2) {
        fprintf(stderr, "ERROR: %s has too few cavities (ncav=%d).\n", ifn, ncav);
        return (1);
    }
    printf("Opened %s:\n", ifn);
    // fetch variables
    rate = sp_var_f4(v, "rate");
    diam = sp_var_f4(v, "diam");
    temp = sp_var_f4(v, "temp");
    attn = sp_var_f4(v, "attn");
    mpsn = sp_var_f4(v, "mpsn");
    npts = sp_var_i2(v, "npts");
    nsrc = (sp_var_find(v, "nsrc") < 0) ? 2 : sp_var_i2(v, "nsrc");
    sets = (sp_var_find(v, "sets") < 0) ? 2 : sp_var_i2(v, "sets");
    if (diam <= 0) {
        diam = 0.8F;
    }
    if (stc > 0) {
        temp = (float) stc;
    }
    // -----------------------------
    printf("npts=%d ncav=%d nsrc=%d ", npts, ncav, nsrc);
    printf("rate=%.0f diam=%0.2f temp=%.0f attn=%.0f\n", rate, diam, temp, attn);
    printf("f_err = %.0f-%.0f; ", erf[0], erf[1]);
    printf("f_hfr = %.0f-%.0f; ", mmf[0], mmf[1]);
    printf("f_zcf = %.0f-%.0f\n", zcf[0], zcf[1]);
    df = rate / npts;   // frequency step
    nf = npts / 2 + 1;  // number of frequencies
    na = nf * 2;        // complex-frequency array size
    // allocate memory for arrays
    pcav = (float *) calloc(nsrc * ncav * na, sizeof(float));
    zcav = (float *) calloc(nsrc * ncav * na, sizeof(float));
    lcav = (float *) calloc(nsrc * (ncav + 1), sizeof(float));
    free_null(psrc);
    free_null(zsrc);
    free_null(pxtk);
    psrc = (float *) calloc(nsrc * na, sizeof(float));
    zsrc = (float *) calloc(nsrc * na, sizeof(float));
    pxtk = (float *) calloc(nsrc * na, sizeof(float));
    // set frequency range for error calculaton
    jef1 = round(erf[0] / df);
    jef2 = round(erf[1] / df);
    // get third-order high-pass filter coefficients
    if (hpf) {
	fco[0] = hpf / rate;
        sp_bessel(B, A, NFC - 1, fco, 1);
    }
    // loop over sound sources
    wrdat = wd;
    itrfl = rr;
    microphone_transfer_init(df, mpsn, nf, 1);
    for (j = 0; j < nsrc; j++) {
        printf("        source #%d: ", j + 1);
        pc = pcav + j * na * ncav;
        zc = zcav + j * na * ncav;
        ps = psrc + j * na;
        zs = zsrc + j * na;
        px = pxtk + j * na;
        lc = lcav + j * (ncav + 1);
        // extract pressure and calculate impedance for each cavity
        for (i = 0; i < ncav; i++) {
            k = i * na;
	    if (cavity_pressure(v, i, j, pc + k)) {
		ncav = i;
		fprintf(stderr, "*** missing data: ncav=%d\n", ncav);
		break;
	    }
            microphone_transfer(pc + k);
	    if (hpf) {
		f = (float *) calloc(nf, sizeof(float));
		h = (float *) calloc(na, sizeof(float));
		sp_linspace(f, nf, 0, (nf - 1) * rate / na);
		sp_freqz(B, NFC, A, NFC, f, h, nf, rate);
		sp_cvmul(h, pc + k, pc + k, nf);
		free(f);
		free(h);
	    }
            cavity_length(pc + k, lc + i);
        }
        dba = db_avg(pc, na, ncav);
        tgd = gd_fpr(pc, na, rate, erf);
        hfr = db_hfr(pc, na, ncav, rate, mmf);
	print_qm(dba, tgd, hfr);
        // estimate Thevenin-equivalent source pressure & impedance
        lc[ncav] = itrfl ? rfl : 0;
        fzero(tau, NTAU);
        compute_cavity_impedance(lc, tau, 0, nf);
        z0 = zc[0];
        thev_source(lc);
        plot_source(j + 1);
        // set frequency range for pc_error calculaton
        jef1 = round(erf[0] / df);
        jef2 = round(erf[1] / df);
        // iterate lengths
        pc_fit(lc, ef1, df);
        pcd = pc_delay(zs, z0);
        if (ewr) {
            tau[0] = (pcd < 0) ? (float)(-pcd / 1000) : 0;
            pc_fit(lc, ef1, df);
            ew_fit(lc, ef2, df);
            pc_fit(lc, ef1, df);
        }
        prn_hw(lc, pcd * 1000);
        // plot impedance for each cavity
        for (i = 0; i < ncav; i++) {
            k = i * na;
            plot_cavity(pc + k, zc + k, i * nsrc + j + 1);
        }
        plot_source(j + 1);
        compute_tube_impedance(zc, lc, tp);
        compute_cavity_impedance(lc, tau, 0, nf);
        thev_source(lc);
        compute_load_impedance(zc, pc);
        srd = reflectance_delay(zsrc, z0, nf, rate, prf);
        if (wrcav) {
            write_cav(ifn, lc, dba, tgd, hfr, srd, pcer, ewer, j);
        }
    }
    // write ps & zs to THS file
    strcpy(tfn, ifn);
    newext(tfn, "THS");
    remove(tfn);
    save_ths(zsrc, psrc, pxtk, lcav, ncav, nsrc, npts, sets,
        rate, df, diam, temp, attn, tfn);
    printf("Wrote %s.\n", tfn);
    if (thl) {
        dot = strchr(tfn, '.');
        strcpy(dot, "N.THL");
        for (i = 0; i < ncav; i++) {
            *dot = 'a' + i;
            pc1 = pcav + i * na;
            pc2 = pc1 + ncav * na;
            save_thl(zsrc, psrc, pxtk, pc1, pc2, na, ncav, nsrc, npts, sets,
                rate, df, attn, tfn);
	    printf("Wrote %s\n", tfn);
	}
    }
    // free memory for arrays
    free(pcav);
    free(zcav);
    free(lcav);

    sp_var_clear_all();     // deallocate variables

    return (0);
}

#ifdef MAKEDLL

int 
GetLength1(VAR *V, double *erf, double *lengths)
{
    double df, z0, pcd;
    int i, j, k, nv, nf, na;
    int16_t sets;

    nv = sp_var_size(V);
   
    // fetch variables
    rate = sp_var_f4(V, "rate");
    diam = sp_var_f4(V, "diam");
    temp = sp_var_f4(V, "temp");
    attn = sp_var_f4(V, "attn");
    mpsn = sp_var_f4(V, "mpsn");
    npts = sp_var_i2(V, "npts");
    ncav = sp_var_i2(V, "ncav");
    nsrc = sp_var_i2(V, "nsrc");
    if(sp_var_find(V,"sets") < 0) {
    	sets = 2;
    } else {
        sets = sp_var_i2(V,"sets");
    }
    if (diam <=0 ) {
        diam = 0.8F;
    }

    df = rate / npts;   // frequency step
    nf = npts / 2 + 1;  // number of frequencies
    na = nf * 2;        // complex-frequency array size
    // allocate memory for arrays
    pcav = (float *) calloc(nsrc * ncav * na, sizeof(float));
    zcav = (float *) calloc(nsrc * ncav * na, sizeof(float));
    lcav = (float *) calloc(nsrc * (ncav + 1), sizeof(float));
    free_null(psrc);
    free_null(zsrc);
    free_null(pxtk);
    psrc = (float *) calloc(nsrc * na, sizeof(float));
    zsrc = (float *) calloc(nsrc * na, sizeof(float));
    pxtk = (float *) calloc(nsrc * na, sizeof(float));
    // loop over sound sources
    wrdat = 0;
    microphone_transfer_init(df, mpsn, nf, 1);
    for (j = 0; j < nsrc; j++) {
        pc = pcav + j * na * ncav;
        zc = zcav + j * na * ncav;
        ps = psrc + j * na;
        zs = zsrc + j * na;
        px = pxtk + j * na;
        lc = lcav + j * (ncav + 1);

        // extract pressure and calculate impedance for each cavity
        for (i = 0; i < ncav; i++) {
            k = i * na;
	    if (cavity_pressure(V, i, j, pc + k)) {
                ncav = i;
                fprintf(stderr, "*** missing data: ncav=%d\n", ncav);
                break;
	    }
            microphone_transfer(pc + k);
            cavity_length(pc + k, lc + i);
        }
        // estimate Thevenin-equivalent source pressure & impedance
        lc[ncav] = itrfl ? rfl : 0;
        fzero(tau, NTAU);
        compute_cavity_impedance(lc, tau, 0, nf);
        z0 = cav0(0, 0, 0);
        thev_source(lc);
        plot_source(j + 1);
        // set frequency range for pc_error calculaton
        jef1 = round(erf[0] / df);
        jef2 = round(erf[1] / df);
        // iterate lengths
        pc_fit(lc, ef1, df);
        pcd = pc_delay(zs, z0);
        if (ewr) {
            tau[0] = (pcd < 0) ? (float)(-pcd / 1000) : 0;
            pc_fit(lc, ef1, df);
            ew_fit(lc, ef2, df);
            pc_fit(lc, ef1, df);
        }
        prn_hw(lc, pcd * 1000);
        // plot impedance for each cavity
        for (i = 0; i < ncav; i++) {
            k = i * na;
            plot_cavity(pc + k, zc + k, i * nsrc + j + 1);
        }
        plot_source(j + 1);
	// pass back error & lengths
        erf[j] = pcer;
        for(i=0; i < ncav; i++)
            lengths[i+ncav*j] = lc[i];
    }
  
    // free memory for arrays
    free(pcav);
    free(zcav);
    free(lcav);

    sp_var_clear_all();     // deallocate variables

    return (0);
}
#endif
