// calthl.c - create THL file from CAL file
#define VER		    "7-Jul-2021"
/* calthl version log:
 * 7-Jul-2021
 * > Report lfp
 * 12-Aug-2020
 * > Tweak to match MATLAB calthl
 * 21-Mar-20f20
 * > Replace long & short with standard types
 * 27-Oct-2016
 * > Added MP_transfer feature
 * 5-Aug-2015
 * > Formatted print output to match calthl.m
 * > Added zl_lfp to print output
 * 30-Apr-2015
 * > Allow THS files created by MATLAB script
 * 23-May-2013
 * > Write zl, pl, pr for option -w
 * 6-May-2011
 * > Fixed bug when "sets" missing from CAL file
 * 20-Apr-2011
 * > Use z_chr as initial value in z_surge
 * 18-Apr-2011
 * > Added phase-rotation
 * 17-Apr-2011
 * > Added to cavcal/VS9
 * 16-Apr-2011
 * > First version
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <sigpro.h>
#include "mptrans.h"
#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#define _access access
#endif

#define MAXLEN        80
#define MAXNCH        8
#define NSURGE        80
#define round(x)      ((int)floor((x)+0.5))
#define free_null(x)  if(x){free(x);x=NULL;}
#define fcopy(x,y,n)    memcpy(x,y,(n)*sizeof(float))
#define limit(a,b,c)    ((b<a)?a:((b>c)?c:b))

static double lmx = 12;     // maximum cavity length (cm)
static double lmn = 1;	    // minimum cavity length (cm)
static double mbw = 20;     // maximum band-width
static double ths_rate;     // source rate
static double ths_diam;     // source cavity diameter
static double ths_temp;     // source cavity temperature
static double ths_attn;	    // source attenuation
static double ths_delf;     // source frequency bin
static double fpr[2] = { 1, 10 };   // primary frequency range
static double fqm[2] = { 0.1, 10 }; // quality-metric frequency range
static double flf[2] = { 0.1, 0.2}; // low frequency range
static double srg[2] = { 0, 0 };
static double ewc[2] = { 0, 0 };
static double dcv[2] = { 0, 0 };
static double lcv[2] = { 0, 0 };
static double zqm[2] = { 0, 0 };
static double lfp[2] = { 0, 0 };
static float *pcav, *zcav;
static float *psrc = NULL;
static float *zsrc = NULL;
static float *pxtk = NULL;
static float *pc, *ps, *zs, *px;
static float rate, diam, temp, attn, mpsn;
static int wrdat = 0;     // write data ?
static int jpf1, jpf2;
static int ncav, nsrc, npts;
static int ths_sets;
static int pbe = 0;                 // pause before exit ?

/***********************************************************************/

// write magnitude & phase (float array)
void
wrdbph(float *z, char *lab, int seq, int unw)
{
    char fn[20];
    double f, df;
    float *db, *ph;
    int  i, np;
    FILE *fp;

    df = rate / npts;
    np = npts / 2;
    db = (float *) calloc(np, sizeof(float));
    ph = (float *) calloc(np, sizeof(float));
    sp_cdb(z, db, np);
    sp_cph(z, ph, np);
    if (unw) {
        sp_unwrap(ph, ph, np);
    }
    sprintf(fn, "%s%02d.txt", lab, seq);
    fp = fopen(fn, "w");
    fprintf(fp, "; %s - from calthl\n", fn);
    fprintf(fp, ";  f        db      ph\n");
    for (i = 0; i < np; i++) {
        f = i * df;
        fprintf(fp, "%6.0f %8.2f %8.3f\n", f, db[i], ph[i]);
    } 
    fclose(fp);
    free(db);
    free(ph);
}

// write magnitude & phase (double array)
void
wrdbphd(double *zd, char *lab, int seq, int unw)
{
    float *zf;
    int  i;

    zf = (float *) calloc(npts, sizeof(float));
    for (i = 0; i < npts; i++) {
        zf[i] = (float) zd[i];
    }
    wrdbph(zf, lab, seq, unw);
    free(zf);
}

// apply new extension to a filename
static void
newext(char *f, char *e)
{
    while (*f != '.' && *f != '\0')
	f++;
    *f++ = '.';
    while (*e != '\0')
	*f++ = *e++;
    *f++ = '\0';
}

/***********************************************************************/

// tube propagation - based on Allen (2020)
double
cav0(double f, double* Z, double* K)
{
    double  c, rho, mu, eta, beta, w, rc, dt, Zchr;
    static double eps = 1e-9;

    rc = diam / 2;      // tube radius
    dt = temp - 26.85;  // temperature deviation
    w = 2 * M_PI * fmax(eps, f);
    c = 3.4723e4 * (1 + 0.00166 * dt);
    rho = 1.1769e-3 * (1 - 0.00335 * dt);
    // propagration function
    if (K) {
        mu = 18.6e-5;
        eta = 1.5345;
        beta = eta * sqrt(mu / rho) / rc;
        K[0] = (beta * sqrt(w / 2)) / c;
        K[1] = (w + beta * sqrt(w / 2)) / c;
    }
    // characteristic impedance
    Zchr = (rho * c) / (M_PI * rc * rc);
    if (Z) {
        Z[0] = Zchr;
        Z[1] = 0;
    }

    return (Zchr);
}

// ideal characterisic impedance
void
z_chr(float *z0)
{
    z0[0] = (float)cav0(0, 0, 0);
    z0[1] = 0;
}

// ideal-cavity impedance
void
z_cav(double f, double lc, float *zc)
{
    double e, c, s, nr, ni, dr, di, dm, ctr, cti, Z[2], G[2];

    if (f == 0) {
	z_chr(zc);
    } else {
	cav0(f, Z, G);
	e = exp(-2 * G[0] * lc);
	c = cos(-2 * G[1] * lc) * e;
	s = sin(-2 * G[1] * lc) * e;
	nr = 1 + c;
	ni = s;
	dr = 1 - c;
	di = -s;
	dm = dr * dr + di * di;
	ctr = (nr * dr + ni * di) / dm;
	cti = (ni * dr - nr * di) / dm;
	// cavity impedance
	zc[0] = (float) (Z[0] * ctr - Z[1] * cti);
	zc[1] = (float) (Z[0] * cti + Z[1] * ctr);
    }
}

/***********************************************************************/

// vector sum
static double 
vsum(double *a, int n)
{
    double sum = 0;
    int i;

    for (i = 0; i < n; i++) {
	sum += a[i];
    }
    return (sum);
}

// vector dot product
static double
vdot(double* a, double* b, int n)
{
    double sum = 0;
    int i;

    for (i = 0; i < n; i++) {
        sum += a[i] * b[i];
    }
    return (sum);
}

/***********************************************************************/

// frequency-domain window
static void
fd_window(double *z, int n, double fdw)
{
    double a, dp, p, w;
    int i, ir, ii;

    a = 0.16;       // Blackman window
    dp = (M_PI / n) / fdw;
    for (i = 1; i < n; i++) {
        ir = i * 2;
        ii = ir + 1;
        p = i * dp;
        if (p < M_PI) {
            w = (1 - a + cos(p) + a * cos(2 * p)) / 2;
        } else {
            w = 0;
        }
        z[ir] *= w;
        z[ii] *= w;
    }
}

// frequency-domain window of real
static void
fd_window_real(double* a, int n, double fdw)
{
    double* aa;
    int i;

    aa = calloc(n * 2, sizeof(double));
    for (i = 0; i < n; i++) {
        aa[i * 2] = a[i];
    }
    fd_window(aa, n, fdw);
    for (i = 0; i < n; i++) {
        a[i] = aa[i * 2];
    }
}

// frequency-domain reflectance
static void 
reflect(double* Gr, double* Gi, double* Zr, double* Zi, double Z0, double* Ze, int n)
{
    double rnr, rni, rdr, rdi,  rdm;
    int i;

    for (i = 0; i < n; i++) {
        rnr = Zr[i] - Z0;
        rdr = Zr[i] + Z0;
        if (Ze) {
            rni = rdi = Zi[i] - Ze[i];
        } else {
            rni = rdi = Zi[i];
        }
        rdm = rdr * rdr + rdi * rdi;
        Gr[i] = (rnr * rdr + rni * rdi) / rdm;
        Gi[i] = (rni * rdr - rnr * rdi) / rdm;
    }
}

// surge impedance
void
z_surge(double* Zl, int na, double* Zxc, double *Ze, double fdw)
{
    double  *Gr, *Gi, * v0, * v1, * v5, *w, *w5, *Zr, *Zi;
    double  Z0, Z1, Z5, sv0, sv1, sv5;
    int     j, k, kr, ki;

    w =  (double*)calloc(na, sizeof(double));
    w5 = (double*)calloc(na, sizeof(double));
    Zr = (double*)calloc(na, sizeof(double));
    Zi = (double*)calloc(na, sizeof(double));
    v0 = (double*)calloc(na, sizeof(double));
    v1 = (double*)calloc(na, sizeof(double));
    v5 = (double*)calloc(na, sizeof(double));
    Gr = (double*)calloc(na, sizeof(double));
    Gi = (double*)calloc(na, sizeof(double));
    for (k = 0; k < na; k++) {
        kr = 2 * k;
        ki = kr + 1;
        w[k] = (2 * M_PI / na) * k;
        w5[k] = pow(w[k], 5);
        v0[k] = 1;
        v1[k] = sin(w[k] / 2);
        v5[k] = sin(w[k] / 10);
        Zr[k] = Zl[kr];
        Zi[k] = Zl[ki];
    }
    fd_window_real(v0, na, fdw);
    fd_window_real(v1, na, fdw);
    fd_window_real(v5, na, fdw);
    sv0 = vsum(v0, na);
    sv1 = vsum(v1, na);
    sv5 = vsum(v5, na);
    for (k = 0; k < na; k++) {
        v0[k] /= sv0;
        v1[k] /= sv1;
        v5[k] /= sv5;
    }
    Z0 = vdot(Zr, v0, na);
    Z1 = fabs(vdot(Zi, w, na) / vdot(w, w, na)) / 3;
    Z5 = Z0 * 4e-4;
    for (k = 0; k < NSURGE; k++) {
        for (j = 0; j < na; j++) {
            Ze[j] = Z1 * w[j] + Z5 * w5[j];
        }
        reflect(Gr, Gi, Zr, Zi, Z0, Ze, na);
        Z0 *= (1 + vdot(Gr, v0, na));
        Z1 *= (1 + vdot(Gi, v1, na));
        Z5 *= (1 + vdot(Gi, v5, na));
    }
    for (k = 0; k < na; k++) {
        Ze[k] = Z1 * w[k] + Z5 * w5[k];
    }
    Zxc[0] = (float)Z0;
    Zxc[1] = (float)(Z1 / Z0);
    Zxc[2] = (float)pow(Z5 / Z0, 0.2);
    free(w);
    free(w5);
    free(Zr);
    free(Zi);
    free(v0);
    free(v1);
    free(v5);
    free(Gr);
    free(Gi);
}

/***********************************************************************/

// ideal characterisic impedance
double
c_sound()
{
    double c, dt;

    dt = temp - 26.85;  // temperature deviation
    c = 3.4723e4 * (1 + 0.00166 * dt);

    return (c);
}

/***********************************************************************/

// transform cavity pressure waveforms to spectra
int
cavity_pressure(VAR *v, int ic, int is, float *pc)
{
    char varnam[20];
    double sc;
    float scale, dacsn, attn, *st, *rs;
    int i, k;
    int32_t *r;
    int16_t *s, sets;

    k = ic * nsrc + is + 1;
    sprintf(varnam, "cal%02d", k);
    if (sp_var_find(v, varnam) < 0) {
	return (1);
    }
    st = (float *) calloc(npts + 2, sizeof(float));
    rs = (float *) calloc(npts + 2, sizeof(float));
    r = v[sp_var_find(v, varnam)].data;
    s = v[sp_var_find(v, "stim")].data;
    for (i = 0; i < npts; i++) {
        rs[i] = (float) r[i];
        st[i] = (float) s[i];
    }
    sp_transfer(st, rs, npts, pc);
    scale = sp_var_f4(v, "scale");
    dacsn = sp_var_f4(v, (is % 2) ? "da2sn" : "da1sn");
    attn = sp_var_f4(v, "attn");
    if (sp_var_find(v, "sets") < 0) {
        sets = 2;
    } else {
        sets = sp_var_i2(v, "sets");
    }
    sc = pow(10, attn / 20) * scale * dacsn / sets;
    sp_smul(pc, pc, npts + 2, sc);
    free(rs);
    free(st);

    return (0);
}

// apply Tukey window
static void
tukey_window(float* pc, float* tw)
{
    float* p;
    int i, i1, i2, i3, i4;

    i1 = round(tw[0] * rate / 1000);
    i2 = round(tw[1] * rate / 1000);
    i3 = round(tw[2] * rate / 1000);
    i4 = round(tw[3] * rate / 1000);
    if (i4 > npts) return;
    p = (float *)calloc(npts + 2, sizeof(float));
    fcopy(p, pc, npts);
    sp_crfft(p, npts);
    for (i = 0; i < npts; i++) {
        if (i < i1) {
            p[i] = 0;
        } else if (i < i2) {
            p[i] *= (1 - (float) cos(M_PI * (i - i1) / (i2 - i1))) / 2;
        } else if (i < i3) {
        } else if (i < i4) {
            p[i] *= (1 + (float) cos(M_PI * (i - i3) / (i4 - i3))) / 2;
        } else {
            p[i] = 0;
        }
    }
    sp_rcfft(p, npts);
    fcopy(pc, p, npts);
    free(p);
}

// estimate cavity length pressure
static void
cavity_length(float* pc, float* lc)
{
    double c, d, t;
    float* p;
    int imn, imx, nf, m;

    p = (float*)calloc(npts + 2, sizeof(float));
    nf = npts / 2 + 1;
    sp_cmagsq(pc, p, nf);
    sp_crfft(p, npts);
    c = c_sound();
    imn = round(lmn * rate * 2 / c);
    imx = round(lmx * rate * 2 / c);
    m = sp_vmax(p + imn, imx - imn) + imn;
    d = (p[m - 1] - p[m + 1]) / (p[m - 1] - 2 * p[m] + p[m + 1]) / 2;
    t = (m + d) / rate;
    *lc = (float)(t * c / 2);
    free(p);
}

// ideal-cavity impedance
void
cavity_impedance(double lc, float *zc)
{
    double df;
    int k, nf;

    nf = npts / 2 + 1;
    df = rate / npts;
    z_chr(zc);  // put z0 at f=0
    for (k = 1; k < nf; k++) {
        z_cav(k * df, lc, zc + k * 2);
    }
}

// compute dB average
double
db_avg(float* pc, int na, int nc)
{
    double p, sum, avg, eps = 1e-20;
    int i, n;

    n = na * nc;
    sum = 0;
    for (i = 0; i < n; i++) {
        p = fabs(pc[i]);
        sum = sum + p * p;
    }
    avg = 2 * sum / n;
    avg = (avg < eps) ? eps : avg;
    return (10 * log10(avg));
}

// compute delay
double
delay(float* p, int nf, double rate)
{
    double ph, dly, f, df, lph = 0;
    double sx = 0, sy = 0, sxx = 0, sxy = 0;
    int i, ir, ii, nav = 0;

    df = rate / (2 * (nf - 1)) / 1000;
    for (i = 0; i < nf; i++) {
        ir = 2 * i;
        ii = ir + 1;
        ph = atan2(p[ii], p[ir]) / (2 * M_PI);
        if (i > 0) { // unwrap phase
            while ((ph - lph) >  0.5) ph -= 1;
            while ((ph - lph) < -0.5) ph += 1;
        }
        lph = ph;
        f = i * df;
        if ((f >= fpr[0]) && (f <= fpr[1])) {
            sx += f;
            sy += ph;
            sxx += f * f;
            sxy += f * ph;
            nav++;
        }
    }
    dly = -(nav * sxy - sx * sy) / (nav * sxx - sx * sx);
    return (dly);
}

// estimate Thevenin-equivalent source pressure & impedance 
void
thev_source(float *lc)
{
    double pir, pii, zir, zii, pms, zms, d, df, f;
    double a0, a1, a2, a3;
    double q0, q1, q2, q3;
    float zc[2];
    int i, ii, ir, k, kr, ki, nc, nf;
    static float eps = 1e-20F;

    nc = ncav;
    //ns = nsrc;
    nf = npts / 2 + 1;
    df = rate / npts;
    // calculate ps & zs
    ps[0] = eps;
    zs[0] = zcav[0];
    px[0] = eps;
    // loop over frequency
    for (k = 1; k < nf; k++) {
        f = k * df;
        // index real & imag
        kr = k * 2;
        ki = kr + 1;
        // zero summation variables
        a0 = a1 = a2 = a3 = 0;
        q0 = q1 = q2 = q3 = 0;
        // loop over cavity
        for (i = 0; i < nc; i++) {
            // index real & imag
            ir = (i * nf + k) * 2;
            ii = ir + 1;
            // get pc & zc for this frq, src, & cav
            pir = pc[ir];
            pii = pc[ii];
            z_cav(f, lc[i], zc);
            zir = zc[0];
            zii = zc[1];
            // sum pc & zc combinations
            pms = pir * pir + pii * pii;
            zms = zir * zir + zii * zii;
            a0 += pms;
            a1 += zms;
            a2 += pir * zir + pii * zii;
            a3 += pii * zir - pir * zii;
            q0 += pir * zms;
            q1 += pii * zms;
            q2 -= zir * pms;
            q3 -= zii * pms;

        }
        // calculate determinant
        d = a1 * a0 - (a2 * a2 + a3 * a3);
        if (fabs(d) < eps) d = eps;
        // solve for ps & zs
        ps[kr] = (float) ((a0 * q0 + a2 * q2 - a3 * q3) / d);
        ps[ki] = (float) ((a0 * q1 + a2 * q3 + a3 * q2) / d);
        zs[kr] = (float) ((a1 * q2 + a2 * q0 + a3 * q1) / d);
        zs[ki] = (float) ((a1 * q3 + a2 * q1 - a3 * q0) / d);
    }
}

/***********************************************************************/

static void
load_ths(const char *tfn, VAR *v)
{
    double *ps1, *ps2, *zs1, *zs2, *px1, *px2, dba[2], dly[2];
    float *ps;
    int i, j, na, np, nf;

    if (sp_var_find(v, "SRC_THV")<0) {
	printf("ERROR: %12s is not a valid THS file.\n", tfn);
	exit(1);
    }
    if (v[0].dtyp == 9) { // created by MATLAB script?
        na = (int) sp_var_f8(v, "SRC_THV");
        ths_sets = (sp_var_find(v, "sets") < 0) ? 2 : (int) sp_var_f8(v, "sets");
    } else {              // EMAV compatible
        na = sp_var_i2(v, "SRC_THV");
        ths_sets = (sp_var_find(v, "sets") < 0) ? 2 : sp_var_i2(v, "sets");
    }
    // fetch variables
    ths_rate = sp_var_f8(v, "rate");
    ths_delf = sp_var_f8(v, "df");
    ths_diam = sp_var_f8(v, "dc");
    ths_temp = sp_var_f8(v, "tc");
    ths_attn = sp_var_f8(v, "attn");
    // fetch arrays
    np = na * 2;
    nf = np / 2 + 1;
    ps1 = v[sp_var_find(v, "ps1")].data;
    ps2 = v[sp_var_find(v, "ps2")].data;
    zs1 = v[sp_var_find(v, "zs1")].data;
    zs2 = v[sp_var_find(v, "zs2")].data;
    px1 = v[sp_var_find(v, "px1")].data;
    px2 = v[sp_var_find(v, "px2")].data;
    free_null(psrc);
    free_null(zsrc);
    free_null(pxtk);
    psrc = (float *) calloc(nf * 4, sizeof(float));
    zsrc = (float *) calloc(nf * 4, sizeof(float));
    pxtk = (float *) calloc(nf * 4, sizeof(float));
    for (i = 0; i < np; i++) {
        j = i + nf * 2;
        psrc[i] = (float) ps1[i];
        psrc[j] = (float) ps2[i];
        zsrc[i] = (float) zs1[i];
        zsrc[j] = (float) zs2[i];
        pxtk[i] = (float) px1[i];
        pxtk[j] = (float) px2[i];
    }
    // -----------------------------
    for (i = 0; i < 2; i++) {
        ps = psrc + i * nf * 2;
        dba[i] = db_avg(ps, na, 1) - 3.01;
        dly[i] = delay(ps, nf, ths_rate);
    }
    printf("%12s: dBavg=%5.1f,%5.1f; delay=%6.2f,%6.2f; attn=%3.0f\n",
        tfn, dba[0], dba[1], dly[0], dly[1], ths_attn);
}

/***********************************************************************/

static void
zl_compute(double* zl, double* pl, double* pr, double* rl, double* zxc, double* ze,
    float* zs, float* ps, float* px, float* pc, int na, int sn)
{
    double  pfr, pfi, plr, pli, plm, prr, pri, psr, psi, por, poi;
    double  dnr, dni, dnm, zsr, zsi, z0, zlr, zli, fdw;
    double* Zsr, * Zsi, * Zlr, * Zli, * Gsr, * Gsi, * Gwr, * Gwi;
    int     i, ii, ir;
    static float eps = 1e-9F;

    Zsr = calloc(na, sizeof(double));
    Zsi = calloc(na, sizeof(double));
    Zlr = calloc(na, sizeof(double));
    Zli = calloc(na, sizeof(double));
    Gsr = calloc(na, sizeof(double));
    Gsi = calloc(na, sizeof(double));
    Gwr = calloc(na, sizeof(double));
    Gwi = calloc(na, sizeof(double));
    zl[0] = pl[0] = pr[0] = eps;
    zl[1] = pl[1] = pr[1] = 0;
    for (i = 1; i < na; i++) {
        ir = 2 * i;
        ii = 2 * i + 1;
        psr = ps[ir];
        psi = ps[ii];
        zsr = zs[ir];
        zsi = zs[ii];
        plr = pc[ir];
        pli = pc[ii];
        // load impedance: Zs*Pl/(Ps-Pl)
        dnr = psr - plr;
        dni = psi - pli;
        dnm = (dnr * dnr) + (dni * dni);
        prr = (plr * dnr + pli * dni) / dnm;
        pri = (pli * dnr - plr * dni) / dnm;
        zlr = zsr * prr - zsi * pri;
        zli = zsi * prr + zsr * pri;
        pl[ir] = plr;
        pl[ii] = pli;
        zl[ir] = zlr;
        zl[ii] = zli;
        // local copy
        Zsr[i] = zsr;
        Zsi[i] = zsi;
        Zlr[i] = zlr;
        Zli[i] = zli;
    }
    fdw = fmin(1, mbw * 2000 / rate);
    z_surge(zl, na, zxc, ze, fdw);
    z0 = zxc[0];
    reflect(Gsr, Gsi, Zsr, Zsi, z0,  0, na); // source reflectance
    reflect(Gwr, Gwi, Zlr, Zli, z0, ze, na); // waveguide reflectance
    for (i = 1; i < na; i++) {
        ir = 2 * i;
        ii = 2 * i + 1;
        // load pressure
        plr = pc[ir];
        pli = pc[ii];
        plm = plr * plr + pli * pli;
        // incident pressure: Po=Ps*(1-Gs)/2
        por = (ps[ir] * (1 - Gsr[i]) - ps[ii] * (0 - Gsi[i])) / 2;
        poi = (ps[ii] * (1 - Gsr[i]) + ps[ir] * (0 - Gsi[i])) / 2;
        // forward pressure: Pf=Po/(1-Gs.*Gw)
        dnr = 1 - (Gsr[i] * Gwr[i] - Gsi[i] * Gwi[i]);
        dni = 0 - (Gsi[i] * Gwr[i] + Gsr[i] * Gwi[i]);
        dnm = (dnr * dnr) + (dni * dni);
        pfr = (por * dnr + poi * dni) / dnm;
        pfi = (poi * dnr - por * dni) / dnm;
        // forward/load pressure ratio: Pf/Pl
        pr[ir] = (pfr * plr + pfi * pli) / plm;
        pr[ii] = (pfi * plr - pfr * pli) / plm;
        // waveguide impedance: Zw=Zo*(1+Gw)./(1-Gw)
        dnr = 1 - Gwr[i];
        dni = 0 - Gwi[i];
        dnm = (dnr * dnr) + (dni * dni);
        zl[ir] = z0 * ((1 + Gwr[i]) * dnr + (0 + Gwi[i]) * dni) / dnm;
        zl[ii] = z0 * ((0 + Gwi[i]) * dnr - (1 + Gwr[i]) * dni) / dnm;
        // waveguide reflectance: Gw
        rl[ir] = Gwr[i];
        rl[ii] = Gwi[i];
    }
    if (wrdat) {
        wrdbphd(zl, "zl", sn, 0);
        wrdbphd(pl, "pl", sn, 1);
        wrdbphd(pr, "pr", sn, 1);
    }
    free(Zsr);
    free(Zsi);
    free(Zlr);
    free(Zli);
    free(Gsr);
    free(Gsi);
    free(Gwr);
    free(Gwi);
}

// estimate cavity diameter from characteristic impedance
static double
cavity_diameter(double Zo)
{
    double c, rho, dt, A, D;

    dt = temp - 26.85;  // temperature deviation
    c = 3.4723e4 * (1 + 0.00166 * dt);
    rho = 1.1769e-3 * (1 - 0.00335 * dt);
    A = rho * c / Zo;
    D = 2 * sqrt(A / M_PI);

    return (D);
}

// estimate cavity length from reflectance 
static double
cavity_length_refl(double* G, int nf, double rate, double* fpr)
{
    double ph, dph, dly, f, df, len, ms;
    double sum1 = 0, sum2 = 0, lph = 0;
    int i, ir, ii;

    df = rate / (2 * (nf - 1)) / 1000;
    for (i = 0; i < nf; i++) {
        ir = 2 * i;
        ii = ir + 1;
        ph = atan2(G[ii], G[ir]) / (2 * M_PI);
        if (i > 0) { // unwrap phase
            while (ph > (lph + 0.5)) ph -= 1;
            while (ph < (lph - 0.5)) ph += 1;
        }
        dph = ph - lph;
        lph = ph;
        f = i * df;
        if ((f > fpr[0]) && (f < fpr[1])) {
            ms = G[ii] * G[ii] + G[ir] * G[ir];
            sum1 += dph * ms; // magnitude-squared weight
            sum2 += ms;
        }
    }
    dly = -(sum1 / sum2) / df;
    len = dly * c_sound() / 2000;
    return (len);
}

static double
zl_info(double* zl, double* pl, double* pr, double* rl, double* zxc, double* ze, int n, int sn)
{
    double Plr, Pli, Pfr, Pfi, Pwr, Pwi, Per, Pei, Pmr, Pmi, Pdr, Pdi;
    double dnr, dni, dnm, zrr, zri, sum1 = 0, sum2 = 0;
    int k, kr, ki, k1, k2;

    k1 = round(2000 * n * fqm[0] / rate);
    k2 = round(2000 * n * fqm[1] / rate);
    for (k = 0; k < n; k++) {
        kr = k * 2;
        ki = kr + 1;
        if ((k >= k1) && (k <= k2)) {
            // load pressure: Pl
            Plr = pl[kr];
            Pli = pl[ki];
            // forward pressure: Pf
            Pfr = Plr * pr[kr] - Pli * pr[ki];
            Pfi = Pli * pr[kr] + Plr * pr[ki];
            // waveguide pressure: Pw=Pf+Pr
            Pwr = Pfr + (Pfr * rl[kr] - Pfi * rl[ki]);
            Pwi = Pfi + (Pfi * rl[kr] + Pfr * rl[ki]);
            // evanescent pressure: Pe=Pw*(Ze/Zw)
            dnr = zl[kr];
            dni = zl[ki];
            dnm = (dnr * dnr) + (dni * dni);
            zrr = (ze[k] * dni) / dnm;
            zri = (ze[k] * dnr) / dnm;
            Per = Pwr * zrr - Pwi * zri;
            Pei = Pwi * zrr + Pwr * zri;
            // model pressure: Pm=Pw+Pe
            Pmr = Pwr + Per;
            Pmi = Pwi + Pei;
            // difference pressure: Pd=Pl-(Pw+Pe)
            Pdr = Plr - Pmr;
            Pdi = Pli - Pmi;
            // accumulate sums
            sum1 += (Pdr * Pdr) + (Pdi * Pdi);
            sum2 += (Plr * Plr) + (Pli * Pli);
        }
    }
    srg[sn - 1] = (float)zxc[0];
    ewc[sn - 1] = (float)(zxc[1] * 2e6 / rate);
    dcv[sn - 1] = (float)cavity_diameter(zxc[0]);
    lcv[sn - 1] = (float)cavity_length_refl(rl, n, rate, fpr);
    zqm[sn - 1] = 1 - sum1 / sum2;
    return (zqm[sn - 1]);
}

static double
zl_lfp(double* zl, int n, int sn)
{
    double df, f1, f2, fk, sum = 0;
    int k, kr, ki, nf = 0;

    f1 = flf[0];
    f2 = flf[1];
    df = rate / (2000 * n);
    for (k = 0; k < n; k++) {
        kr = k * 2;
        ki = kr + 1;
        fk = k * df;
        if ((fk >= f1) && (fk <= f2)) {
            sum += atan2(zl[ki], zl[kr]) / (2 * M_PI);
            nf++;
        }
    }
    lfp[sn - 1] = sum / nf;
    return (lfp[sn - 1]);
}

static void
irdft(double* ft, int n, int nc, int ic, double fdw)
{
    double dp;
    int j, jn, k, kr, ki;

    dp = 2 * M_PI / n;
    for (j = 0; j < nc; j++) {
        jn = j * n * 2;
        for (k = 0; k < n; k++) {
            kr = k * 2;
            ki = kr + 1;
            ft[jn + kr] = cos(dp * ic * j) / n;
            ft[jn + ki] = sin(dp * ic * j) / n;
        }
        fd_window(ft + jn, n, fdw);
        ic++;
    }
}

static void
zl_repair(double* zl, double* zxc, int n)
{
    double a, b, df, fdw, zr, zi, zrmin, zimin, zmax, zm, ym, yr, yi;
    double* zrlim, * zilim, *ft;
    double zrsum = 0, zisum = 0, sum = 0;
    int j, jn, k, kr, ki, k1, k2, ic, nc;
    static double f1 = 80;  // very-low frequency
    static double f2 = 200; // low frequency

    zrlim = calloc(n, sizeof(double));
    zilim = calloc(n, sizeof(double));
    df = rate / (2 * n);
    // low frequency repair
    k1 = round(f1 / df);
    k2 = round(f2 / df);
    zmax = zxc[0] * 10;
    zrmin = zxc[0] / 100;
    zimin = -zmax;
    for (k = 0; k < k2; k++) {
        kr = k * 2;
        ki = kr + 1;
        zrlim[k] += limit(zrmin, zl[kr], zmax);
        zilim[k] += limit(zimin, zl[ki], zmax);
        zrsum += zrlim[k];
        zisum += zilim[k];
    }
    zr = zrsum / k2;
    zi = zisum / k2;
    for (k = 0; k < k1; k++) {
        kr = k * 2;
        ki = kr + 1;
        a = (1 + cos(M_PI * k / k1)) / 2;
        b = (1 - cos(M_PI * k / k1)) / 2;
        zl[kr] = a * zr + b * zrlim[k];
        zl[ki] = a * zi + b * zilim[k];
    }
    // flip negative-real admittance
    for (k = 0; k < k1; k++) {
        kr = k * 2;
        ki = kr + 1;
        zm = (zl[kr] * zl[kr]) + (zl[ki] * zl[ki]);
        yr = zl[kr] / zm;
        yi = -zl[ki] / zm;
        if (yr < 0) {
            ym = (yr * yr) + (yi * yi);
            zl[kr] = -yr / ym;
            zl[ki] = -yi / ym;
        }
    }
    // fix zero-frequency component
    ic = -16; // initial component
    nc = 9;   // number of components
    fdw = fmin(1, mbw * 2000 / rate);
    ft = calloc(n * nc * 2, sizeof(double));
    irdft(ft, n, nc, ic, fdw);
    for (j = 0; j < nc; j++) {
        jn = j * n * 2;
        sum = 0;
        for (k = 1; k < n; k++) {
            kr = k * 2;
            ki = kr + 1;
            sum += zl[kr] * ft[jn + kr];
            sum += zl[ki] * ft[jn + ki];
        }
    }
    free(ft);
    zl[0] = -sum / nc;
    zl[1] = 0;
}

static void
save_thl(float* zs1, float* ps1, float* px1, float* pc1, float* pc2,
    int na, int nc, int ns, int np, int16_t sets,
    double rate, double df, double at, const char* tfn)
{
    double* pl1, * pl2, * pr1, * pr2, * rl1, * rl2, * zl1, * zl2, * ze1, * ze2, * z0;
    double zxc1[3], zxc2[3], qm1, qm2;
    float* zs2, * ps2, * px2, zc[2];
    int i, ir, ii, n, nv;
    VAR* vl;
    static char ftype[] = "LD_THV";

    // convert arrays from float to double
    z0 = (double*)calloc(np, sizeof(double));
    zl1 = (double*)calloc(np, sizeof(double));
    zl2 = (double*)calloc(np, sizeof(double));
    ze1 = (double*)calloc(np, sizeof(double));
    ze2 = (double*)calloc(np, sizeof(double));
    pl1 = (double*)calloc(np, sizeof(double));
    pl2 = (double*)calloc(np, sizeof(double));
    pr1 = (double*)calloc(np, sizeof(double));
    pr2 = (double*)calloc(np, sizeof(double));
    rl1 = (double*)calloc(np, sizeof(double));
    rl2 = (double*)calloc(np, sizeof(double));
    z_chr(zc);
    n = np / 2;
    for (i = 0; i < n; i++) {
        ir = i * 2;
        ii = ir + 1;
        z0[ir] = zc[0];
        z0[ii] = zc[1];
    }
    zs2 = zs1 + na;
    ps2 = ps1 + na;
    px2 = px1 + na;
    zl_compute(zl1, pl1, pr1, rl1, zxc1, ze1, zs1, ps1, px1, pc1, n, 1);
    zl_compute(zl2, pl2, pr2, rl2, zxc2, ze2, zs2, ps2, px2, pc2, n, 2);
    qm1 = zl_info(zl1, pl1, pr1, rl1, zxc1, ze1, n, 1);
    qm2 = zl_info(zl2, pl2, pr2, rl2, zxc2, ze2, n, 2);
    zl_lfp(zl1, n, 1);
    zl_lfp(zl2, n, 2);
    zl_repair(zl1, zxc1, n);
    zl_repair(zl2, zxc2, n);
    // set variables
    nv = 14;                // number of variables
    vl = sp_var_alloc(nv);
    sp_var_set(vl + 0, ftype, &n, 1, 1, "i2");
    sp_var_set(vl + 1, "rate", &rate, 1, 1, "f8");
    sp_var_set(vl + 2, "df", &df, 1, 1, "f8");
    sp_var_set(vl + 3, "z0", z0, n, 1, "f8c");
    sp_var_set(vl + 4, "zl1", zl1, n, 1, "f8c");
    sp_var_set(vl + 5, "pl1", pl1, n, 1, "f8c");
    sp_var_set(vl + 6, "pr1", pr1, n, 1, "f8c");
    sp_var_set(vl + 7, "zl2", zl2, n, 1, "f8c");
    sp_var_set(vl + 8, "pl2", pl2, n, 1, "f8c");
    sp_var_set(vl + 9, "pr2", pr2, n, 1, "f8c");
    sp_var_set(vl + 10, "attn", &at, 1, 1, "f8");
    sp_var_set(vl + 11, "sets", &sets, 1, 1, "i2");
    sp_var_set(vl + 12, "qm1", &qm1, 1, 1, "f8");
    sp_var_set(vl + 13, "qm2", &qm2, 1, 1, "f8");
    // save file & clean-up
    sp_mat_save(tfn, vl);   // save variables in MAT file
    sp_var_clear(vl);
    free(z0);
    free(zl1);
    free(zl2);
    free(ze1);
    free(ze2);
    free(pl1);
    free(pl2);
    free(pr1);
    free(pr2);
}

/***********************************************************************/

int
calthl(char *ifn, int wd, int rph, double *phf)
{
    char tfn[MAXLEN];
    double df, dba[2], dly[2];
    float *lc, *pc1, *pc2, lcav[MAXNCH * 2];
    int i, j, k, nf, na;
    int16_t sets;
    VAR *v;
    static char ft_ths[] = "SRC_THV";
    static char ft_cal[] = "EMAV_CAL";
    static float tw[4] = { 0, 1, 61, 62 };

    // open CAL file
    v = sp_mat_load(ifn);
    if (!v) {
        fprintf(stderr, "ERROR: can't open %s.\n", ifn);
        return (1);
    }
    if (strcmp(v[0].name, ft_ths) == 0) {
        load_ths(ifn, v);
	sp_var_clear_all();     // deallocate variables
        return (0);
    } else if (strcmp(v[0].name, ft_cal) != 0) {
        fprintf(stderr, "ERROR: %s file type is not %s or %s.\n",
	    ifn, ft_ths, ft_cal);
	sp_var_clear_all();     // deallocate variables
        return (1);
    }
    ncav = sp_var_i2(v, "ncav");
    if (ncav < 1) {
        fprintf(stderr, "ERROR: %s has too few cavities (ncav=%d).\n",
	    ifn, ncav);
	sp_var_clear_all();     // deallocate variables
        return (1);
    } else if (ncav > 1) {
        fprintf(stderr, "ERROR: %s has too many cavities (ncav=%d).\n",
	    ifn, ncav);
	sp_var_clear_all();     // deallocate variables
        return (1);
    } else if (psrc == NULL) {
        fprintf(stderr, "ERROR: THS file must precede CAL files.\n");
	sp_var_clear_all();     // deallocate variables
        return (1);
    }
    // fetch variables
    rate = sp_var_f4(v, "rate");
    attn = sp_var_f4(v, "attn");
    mpsn = sp_var_f4(v, "mpsn");
    npts = sp_var_i2(v, "npts");
    nsrc = sp_var_i2(v, "nsrc");
    if (sp_var_find(v, "sets") < 0) {
        sets = 2;
    } else {
        sets = sp_var_i2(v, "sets");
    }
    // -----------------------------
    df = rate / npts;   // frequency step
    nf = npts / 2 + 1;  // number of frequencies
    na = nf * 2;        // complex-frequency array size
    // allocate memory for arrays
    pcav = (float *) calloc(na * nsrc * ncav, sizeof(float));
    zcav = (float *) calloc(na * nsrc, sizeof(float));
    // set frequency range for phase calculaton
    if (rph) {
        jpf1 = round(phf[0] / df);
        jpf2 = round(phf[1] / df);
    } else {
        jpf1 = jpf2 = 0;
    }
    // loop over sound sources
    wrdat = wd;
    microphone_transfer_init(df, mpsn, nf, 1);
    for (j = 0; j < nsrc; j++) {
        pc = pcav + j * na * ncav;
        ps = psrc + j * na;
        zs = zsrc + j * na;
        px = pxtk + j * na;
        lc = lcav + j * ncav;
        for (i = 0; i < ncav; i++) {
            k = i * na;
	    if (cavity_pressure(v, i, j, pc + k)) {
		ncav = i;
		fprintf(stderr, "*** missing data: ncav=%d\n", ncav);
		break;
	    }
            microphone_transfer(pc + k);
            tukey_window(pc + k, tw);
            cavity_length(pc + k, lc + i);
        }
        dba[j] = db_avg(pc, na, ncav);
        dly[j] = delay(pc, nf, rate);
        lcv[j] = lc[0];
    }
    strncpy(tfn, ifn, MAXLEN);
    newext(tfn, "THL");
    remove(tfn);
    // Assume ncav=1
    temp = 25;
    pc1 = pcav;
    pc2 = pcav + ncav * na;
    save_thl(zsrc, psrc, pxtk, pc1, pc2, na, ncav, 
        nsrc, npts, sets, rate, df, attn, tfn);
    printf("%14s: dBavg=%6.1f,%6.1f; delay=%6.2f,%6.2f; ", ifn, dba[0], dba[1], dly[0], dly[1]);
    printf("attn=%3.0f; ewc=%6.2f,%6.2f\n", ths_attn, ewc[0], ewc[1]);
    printf("%14s: dcav=%5.2f,%5.2f; zchr=%5.1f,%5.1f; ", tfn, dcv[0], dcv[1], srg[0], srg[1]);
    printf("lcav=%5.2f,%5.2f; zqm =%5.2f,%5.2f; ", lcv[0], lcv[1], zqm[0], zqm[1]);
    printf("lfp =%5.2f,%5.2f\n", lfp[0], lfp[1]);

    // free memory for arrays
    free(pcav);
    free(zcav);

    sp_var_clear_all();     // deallocate variables

    return (0);
}

static void
done()
{
    if (pbe) {
        printf("Press Enter...\n");
	getchar();
    }
    exit(0);
}

static void
usage()
{
    printf("\n");
    printf("usage: calthl [-options] THSfile CALfile ...\n");
    printf("\n");
    printf("option:\n");
    printf(" -p    - pause before exit\n");
    printf(" -w    - create TXT files for plotting data\n");
    printf("\n");
    done();
 }

int
main(int ac, char **av)
{
    double phf[2];
    static int wd = 0;                  // write data ?
    static int rph = 0;                 // rotate phase ?

    phf[0] = 1000;  // min. phase-rotation frequency (Hz)
    phf[1] = 9000;  // max. phase-rotation frequency (Hz)

    printf("---- calthl %s ----\n", VER);

    // parse command-line
    if (ac < 2) {
        usage();
    }
    if (ac < 3) {
        usage();
    }
    while (ac > 1) {
        if (_access(av[1],0) == 0) {    // does file exist ?
            if (calthl(av[1], wd, rph, phf)) {    // create THL file
		return (1);
            }
            if (ac < 3) {
		done();
            }
        } else if (av[1][0] == '-') {   // options ?
	    if (av[1][1] == 'f' && av[1][2] == '1') {
                phf[0] = atof(av[2]);               // min. phase-rotation freq.
	    } else if (av[1][1] == 'f' && av[1][2] == '2') {
                phf[1] = atof(av[2]);               // max. phase-rotation freq.
	    } else if (av[1][1] == 'p') {
                pbe = 1;		// pause before exit
	    } else if (av[1][1] == 'w') {
                wd = 1;			// write data
	    } else {
		usage();
	    }
	    // bump options with arguments
	    if (strchr("ffz",av[1][1])) {
		ac--;
		av++;
	    }
        } else {
            usage();
        } 
        ac--;
        av++;
    }

    return (0);
}
