// cavcal.c - perform cavity pressure measurements
#define VER		    "21-Jun-2021"
/* cavcal version log:
 * 21-Jun-2021
 * > Changed "int" to "int32_t" in CAV_USAGE structure
 * 12-Aug-2020
 * > Reduced number of parameters to NTAU=2 in evanescent-wave model
 * > Tweak to match MATLAB calthl
* 21-Mar-2020
 * > Replace long & short with standard types
* 9-Nov-2016
 * > remove code that puts "MP_transer" into environment
* 27-Oct-2016
 * > Added MP_transfer feature
 * 17-Oct-2016
 * > Increase size of mp_trans to 256
 * > Sets environment variable "MP_transfer"
 * 14-Jun-2016
 * > revert to version 12-Dec-2016
 * > update CAV_USAGE struct
 * > update Init() function
 * > copy calsrc.c from version 19-May-2016
 * > copy prbths.c from version 19-May-2016
 * > update main() function
 * > set na=nrep*2 in get_response()
 * 19-May-2016
 * > removed response factor of 2 in get_response to restore 29-Feb-2016 version
 * > added stimulus factor of 2 in get_response
 * 4-Apr-2016
 * > get_response scale factor multiplied by 2 for EMAV compatibility
 * 29-Feb-2016
 * > use SIO functions for get_response
 * 25-Feb-2016
 * > Added mp & mp_trans to CAV_USAGE struct
 * > Added GetResponse function to DLL
 * 9-Feb-2016
 * > Set _reps=_nrep/(2*_sets) in Init for cavcal.dll
 * 21-Jan-2016
 * > Modified zc_fit parameter scale factor
 * 19-Jan-2016
 * > Corrected loop error in print_hom
 * 12-Dec-2015
 * > Added cavcal_version to cavcal.dll
 * 2-Dec-2015
 * > Added decimal place to zc_lfp
 * 1-Dec-2015
 * > Cleaned up for MinGW
 * > Corrected error in zc_lfp
 * 30-Nov-2015
 * > Added quality metric zc_lfp (impedance low-frequency phase)
 * 20-Nov-2015
 * > Fixed DLL compile errors
 * > Fixed lcav allocation
* 19-Nov-2015
 * > Added zc_fit & zc_error
 * 17-Nov-2015
 * > Cleaned up memory allocation for DLL
 * 12-Nov-2015
 * > Sync calsrc.c with prbths.c
 * > Fixed memory allocation in DLL for Stepper
 * 30-Sep-2015
 * > print arsc version
 * 28-Sep-2015
 * > Fix parsgin of -a argument
 * 2-May-2015
 * > Make dacsn optional in PRB file
 * 30-Apr-2015
 * > Revisions to prbths & calthl only
 * 6-Jan-2015
 * > Removed factor of 4 in cavity_pressure
 * 23-May-2013
 * > Write zl, pl, pr for option -w
 * 14-May-2013
 * > Move get_vfs out of save_cal
 * 18-Feb-2012
 * > Avoid ideal-reflection > 1 with option -R
 * 2-May-2011
 * > Added option -R to disable ideal-reflection reduction
 * 22-Apr-2011
 * > Removed option for setting ideal-cavity reflection
 * > Added ideal-cavity reflection to length iteration
 * 21-Apr-2011
 * > Added option for setting ideal-cavity reflection
 * 20-Apr-2011
 * > Use z_chr as initial value in z_surge
 * 18-Apr-2011
 * > Fixed high-pass filter in SIGPRO
 * 17-Apr-2011
 * > Replaced freqz & cvmul with SIGPRO functions
 * 16-Apr-2011
 * > Set minimum value in surge()
 * > Add high-pass filter option
 * 09-Feb-2011
 * > Added "surge" impedance to THL files.
 * 01-Feb-2011
 * > Added db_avg print to calsrc
 * 16-Dec-2010
 * > Fixed -t to specify temperature in CAL file
 * > Allow -t to override temperature in CAL file
 * 18-Nov-2010
 * > Scaled THS & THL files to match EMAV v321
 * 29-Oct-2010
 * > Fixed bug reading "sets"
 * 28-Oct-2010
 * > Appended "sets" to THS/THL files
 * 26-Oct-2010
 * > Added command-line option for "sets"
 * > Defined reps as sweeps_per_set
 * > Calculated nrep = reps * sets * 2;
 * 26-Oct-2010
 * > Changed variable name "nrep" to "reps" in CAL/PRB file
 * > Appended "sets" to CAL/PRB files
 * 25-Oct-2010
 * > Removed "* 2" in cavity_pressure()
 * 19-Oct-2010
 * > Replace diam=0 in CAL/PRB file with diam=0.8
 * > Make debug print standard
 * 2-Oct-2010
 * > Fixed memory leak in cavity_length
 * 1-Oct-2010
 * > Improved initial guess of cavity_length
 * 30-Sep-2010
 * > fixed scaling bug in cavity_pressure
 * 16-Sep-2010
 * > Added ar_io_open call to save_cal
 * 27-Aug-2010
 * > Added half-wave frequencies
 * 26-Aug-2010
 * > Added -m (mp sensitivity) option
 * 25-Aug-2010
 * > Added error frequency range (erf)
 * > Added length-iteration options
 * > Added pause-before-exit option
 * > Fixed white-noise-stimulus option
 * 10-Aug-2010
 * > Make z0(0) nonzero
 * 9-Aug-2010
 * > Make pr(0) nonzero
 * > Correct numbering of TXT files
 * 8-Aug-2010
 * > Added option to write THL files
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <arsclib.h>
#include <sigpro.h>
#ifdef MAKEDLL
#include "cavcaldll.h"
#endif
#ifdef WIN32
#include <io.h>
#include <Windows.h>
#define msleep(x) Sleep(x)
#else
#include <unistd.h>
#define _access access
#define msleep(x)  usleep((x)*1000)
#endif

#define MAXLEN              80
#define MAXNCH              8
#define limit(min,val,max)  ((val<min)?min:((val>max)?max:val))
#define round(x)            ((int)floor((x)+0.5))
#define free_null(x)        if(x){free(x);x=NULL;}

int calsrc(char *, double *, int, int, int, int, int);
#ifdef MAKEDLL
int stepper_init();
int stepper_moveto(int);
int stepper_close();
int GetLength1(VAR *,double *,double *);
/*
Return Codes: 
0  Normal Operation
-1 stepper not present
-2 soundcard not present
*/
typedef struct {
    int32_t at;         // attenuation
    int32_t npts;       // buffer size
    int32_t thl;        // thl flag
    int32_t styp;       // white-noise stimulus
    int32_t fs;         // sampling rate
    int32_t nrep;       // number of sweeps
    int32_t wd;         // write data ?
    int32_t nCav;       // number of cavities to test
    int32_t ths;        // ths flag. write ths file
    double  dc;         // cavity diameter
    double  mp;         // microphone sensitivity
    wchar_t mp_trans[256]; // microphone transfer-function filename
} CAV_USAGE;
#else
#define _FUNC(x) x
#endif

static void
gen_sweep(float *s, int n)
{
    double df, ph;
    int i;

    df = M_PI / n;
    ph = 0;
    for (i = 0; i < n; i++) {
	ph += i * df;
	s[i] = (float) sin(ph);
    }
}

static void
gen_stim(float *st, int npts, int type)
{
    double sc;
    int i;

    if (type) {
        gen_sweep(st, npts);     // chirp stimulus
    } else {
        sp_randflat(st, npts);   // white-noise stimulus
    }
    sc = pow(2, 15) - 1;         // maximum positive 16-bit integer
    for (i = 0; i < npts; i++) {
        st[i] = (float) round(st[i] * sc);
    }
}

static void
detrend(float *s, int n)
{
    double  x, y, sx, sy, sxy, sxx, syy, a0, a1;
    int     i;

    sx = sy = sxy = sxx = syy = 0;
    for (i = 0; i < n; i++) {
	x = i;
	y = s[i];
	sx += x;
	sy += y;
	sxy += x * y;
	sxx += x * x;
	syy += y * y;
    }
    a1 = (sxy - sx * sy / n) / ( sxx - sx * sx / n);
    a0 = (sy - a1 * sx) / n;
    for (i = 0; i < n; i++)
	s[i] -= (float)(a0 + a1 * i);
}

int
genfn(const char *c, const char *e, char *f)
{
    char   *a, *b;
    int     i, y, m, d;
    time_t  clk;
    struct tm *tm_;

    // get current date
    clk = time(&clk);
    tm_ = localtime(&clk);
    y = tm_->tm_year % 100;
    m = tm_->tm_mon;
    d = tm_->tm_mday;
    // generate file name with date
    sprintf(f, "%02d%c%02d%.1s00.%s", y, m + 'A', d, c, e);
    // insert unique sequence number
    a = f + 6;
    b = a + 1;
    for (i = 0; i < 100; i++) {
        *a = i / 10 + '0';
        *b = i % 10 + '0';
	if (_access(f, 0) != 0) {
	    return (0);
	}
    }
    *a = '#';
    *b = '#';
    return (1);
}

static void
print_error(char *msg)
{
    fprintf(stderr, "ERROR: %s\n", msg);
}

static int
check_accept(int *x, int *y, int n, int cnt)
{
    // implement "artifact rejection" here
    // return 0 to reject sweep
    return (1);
}

static int
check_rule(int *x, int n, int cnt)
{
    // implement "measurement-based stopping rules" here
    // return 1 to stop averaging
    return (0);
}

static int
check_user()
{
    // allow user to terminate averaging immediately
    // return 1 to stop averaging
    return (0);
}

static int
sync_avg(int id, float *x, float *y, double at,
    int npnt, int nswp, int nskp, int co, int ci)
{
    char msg[MAXLEN];
    double sc;
    int nchn = 2;
    int i, j, n, cur, cnt, prv, err, nseg, *ix[2], *iy[2], *jx, *jy;
    int32_t *siz; 
    void **inn, **out; 
    static int nbuf = 2; 			/* double buffer */
    static int32_t fmt[2] = {ARSC_DATA_I4, 0};	/* long,non-interleaved */

    n = npnt;
    memset(y, 0, n * sizeof(float));	/* zero response */
    if (nswp < 1 || nchn < 1) {
	return (0);
    }
    /* configure stimulus & response buffers */
    for (i = 0; i < nbuf; i++) {
	ix[i] = (int *) calloc(n, sizeof(int));
	iy[i] = (int *) calloc(n, sizeof(int));
    }
    jx = ix[0];
    sc = pow(2, 16) * pow(10, -at / 20);
    for (j = 0; j < npnt; j++) {
        jx[j] = round(x[j] * sc);
    }
    /* configure i/o buffers */
    nseg = nswp + nskp + 1;
    inn = (void **) calloc(nseg * nchn, sizeof(void *));
    out = (void **) calloc(nseg * nchn, sizeof(void *));
    siz = (int32_t *) calloc(nseg, sizeof(int32_t));
    for (i = 0; i < nseg; i++) {
	out[i * nchn + co] = (i < (nswp + nskp)) ? ix[0] : NULL;
	inn[i * nchn + ci] = iy[i % nbuf];
	siz[i] = npnt;
    }
    /* initiate i/o */
    err = ar_set_fmt(id, fmt);
    if (err) {
	ar_err_msg(err, msg, MAXLEN);
	print_error(msg);
	return (0);
    }
    err = ar_io_prep(id, inn, out, siz, nseg, nseg);
    if (err) {
	ar_err_msg(err, msg, MAXLEN);
	print_error(msg);
	return (0);
    }
    err = ar_io_start(id);
    if (err) {
	ar_err_msg(err, msg, MAXLEN);
	print_error(msg);
	return (0);
    }
    jx = ix[1];
    cur = cnt = prv = 0;
    while (cur < nseg && cnt < nswp) {
	cur = ar_io_cur_seg(id);
	if (cur != prv && cur > nskp && cnt < nswp) {
	    jy = iy[(cur - 1) % nbuf];
	    if (check_accept(jx, jy, n, cnt)) {
                for (j = 0; j < n; j++) {
                    jx[j] += jy[j] >> 8;
                }
	        cnt++;
	    }
	    prv = cur;
            if (check_rule(jx, n, cnt)) {
                nswp = cnt;
	    }
	}
        if (check_user()) {
              break;
	}
    }
    /* fetch response */
    for (j = 0; j < npnt; j++) {
        y[j] = (float) jx[j];
    }
    /* de-allocate buffers */
    for (i = 0; i < nbuf; i++) {
	free(ix[i]);
	free(iy[i]);
    }
    free(inn);
    free(out);
    free(siz);

    return (cnt);
} 

static int
get_response(int np, int nr, int co, int ci, int sk, double fs, double at,
    int id, float *st, float *rs)
{
    int cnt, na, nch = 2;

    na = nr * 2;
    ar_io_open(id, fs, nch, nch);
    cnt = sync_avg(id, st, rs, at, np, na, sk, co, ci);
    ar_io_close(id);

    return (cnt);
}

static void
freq_resp(float *H, int nf, double df, char *fn)
{
    double f;
    float *db, *ph, *gd;
    int i;
    FILE *fp;
    
    db = (float *) calloc(nf, sizeof(float));
    ph = (float *) calloc(nf, sizeof(float));
    gd = (float *) calloc(nf, sizeof(float));
    sp_cdb(H, db, nf);          // decibels
    sp_cph(H, ph, nf);          // phase
    sp_unwrap(ph, ph, nf);      // unwrap phase
    sp_cgd(H, gd, nf, df);      // delay
    
    fp = fopen(fn, "w");
    if (!fp) {
	printf("can't open %s\n", fn);
	return;
    }
    fprintf(fp, "; %s - transfer-function test test\n\n", fn);
    fprintf(fp, ";  f        dB       ph       gd\n");
    for (i = 0; i < nf; i++) {
	f = i * df;
	fprintf(fp, "%8.3f %8.3f %8.3f %8.3f\n", f, db[i], ph[i], gd[i]);
    }
    fclose(fp);
    free(db);
    free(ph);
    free(gd);
}

_FUNC(void) cavcal_version(char *s1, char *s2, char *s3)
{
    if (s1)
        sprintf(s1, "CavCal version %s", VER);
    if (s2)
        sprintf(s2, "%s", ar_version());
    if (s3)
        sprintf(s3, "%s", sp_version());
}

#ifdef MAKEDLL
_FUNC(void) save_cal_nosoundcardwithdifference(float *st, float *rs,float * difs,
    double fs, double at, double mp, double dc, double tc, double sc,  double ac, double ds1, double ds2,
    int nr, int ns, int np, int nc, int no, int ni, 
    char *sfn,  char *cfn)
{
    char wvn[8];
	char dif[8];

   
    float rate, attn, adsn, da1sn, da2sn, mpsn, scale;
    float diam, temp;
    int i, j, sz, nch = 2;
    int32_t *s4;
    int32_t * s5;
    int16_t *s2;
    int16_t nv, nw, reps, npts, ncav, nsrc, nadc, sets;
    VAR *vl;
    static char ftype[] = "EMAV_CAL";
	
    rate = (float)fs;           // sampling rate
    npts = (int16_t)np;           // number of points
    reps = (int16_t)nr;           // sweeps per set
    attn = (float)at;           // stimulus attenuation
    mpsn = (float)mp;           // microphone sensitivity
    ncav = (int16_t)nc;           // number of cavities
    nsrc = (int16_t)no;           // number of sources
    nadc = (int16_t)ni;           // number of A/D converters
    diam = (float)dc;           // cavity diameter (cm)
    temp = (float)tc;           // cavity temperature (C)
    scale = (float)sc;          // response waveform scale factor
    sets = (int16_t)ns;           // number of sweep sets
   
	adsn = (float)ac;
    da1sn = (float)ds1;
    da2sn = (float)ds2;   	

    // allocate memory
    s2 = (int16_t *)calloc(npts, sizeof(int16_t));
    s4 = (int32_t *) calloc(npts, sizeof(int32_t));
    s5 = (int32_t *) calloc(npts, sizeof(int32_t));

    // convert stimulus waveform to short array
    for (i = 0; i < npts; i++) {
        s2[i] = (int16_t)round(st[i]);
    }

    // set variables
    nw = nc * no * ni;          // number of waveforms
    nv =  nw+19+nw;             // number of variables // added one for dif
    sz =strlen(sfn);            // length of stimulus name

    vl = sp_var_alloc(nv);

    sp_var_set(vl + 0, ftype,  &nw,   1, 1, "i2");
    sp_var_set(vl + 1, "rate", &rate, 1, 1, "f4");
    sp_var_set(vl + 2, "npts", &npts, 1, 1, "i2");
    sp_var_set(vl + 3, "reps", &reps, 1, 1, "i2");
    sp_var_set(vl + 4, "attn", &attn, 1, 1, "f4");
    sp_var_set(vl + 5, "adsn", &adsn, 1, 1, "f4");
    sp_var_set(vl + 6, "da1sn", &da1sn, 1, 1, "f4");
    sp_var_set(vl + 7, "da2sn", &da2sn, 1, 1, "f4");
    sp_var_set(vl + 8, "mpsn", &mpsn, 1, 1, "f4");
    sp_var_set(vl + 9, "scale", &scale, 1, 1, "f4");
    sp_var_set(vl + 10, "stmfn", sfn, 1, sz, "t");
    sp_var_set(vl + 11, "stim", s2, npts, 1, "i2");
    sp_var_set(vl + 12, "ncav", &ncav, 1, 1, "i2");
    sp_var_set(vl + 13, "nsrc", &nsrc, 1, 1, "i2");
    sp_var_set(vl + 14, "nadc", &nadc, 1, 1, "i2");
    sp_var_set(vl + 15, "diam", &diam, 1, 1, "f4");
    sp_var_set(vl + 16, "temp", &temp, 1, 1, "f4");
    // convert response waveforms to long arrays
    for (j = 0; j < nw; j++) {
        sprintf(wvn, "cal%02d", j + 1);
        for (i = 0; i < npts; i++) {
           s4[i] = (int32_t)round(rs[i + j * npts]);
        }
        sp_var_set(vl + 17 + j, wvn, s4, npts, 1, "i4");
    }

////////////////////////////////////////////////////////////////

    for (j = 0; j < nw; j++) {
        sprintf(dif, "dif%02d", j + 1);
        for (i = 0; i < npts; i++) {
           s5[i] = (int32_t)round(difs[i + j * npts]);
        }
        sp_var_set(vl + 18 + nw + j, dif, s5, npts, 1, "i4");
    }

////////////////////////////////////////////////////////////////
    sp_var_set(vl + 18 +  nw +j , "sets", &sets, 1, 1, "i2");
    // save file & clean-up
    sp_mat_save(cfn, vl);   // save variables in MAT file
    sp_var_clear_all();     // deallocate variables
    // free memory
    free(s2);
    free(s4);
    free(s5);	
}

_FUNC(void) save_cal_nosoundcard(float *st, float *rs,
    double fs, double at, double mp, double dc, double tc, double sc,  double ac, double ds1, double ds2,
    int nr, int ns, int np, int nc, int no, int ni, 
    char *sfn,  char *cfn)
{
    char wvn[8];
    float rate, attn, adsn, da1sn, da2sn, mpsn, scale;
    float diam, temp;
    int i, j, sz, nch = 2;
    int32_t *s4;
    int16_t *s2;
    int16_t nv, nw, reps, npts, ncav, nsrc, nadc, sets;
    VAR *vl;
    static char ftype[] = "EMAV_CAL";
	
    rate = (float)fs;           // sampling rate
    npts = (int16_t)np;           // number of points
    reps = (int16_t)nr;           // sweeps per set
    attn = (float)at;           // stimulus attenuation
    mpsn = (float)mp;           // microphone sensitivity
    ncav = (int16_t)nc;           // number of cavities
    nsrc = (int16_t)no;           // number of sources
    nadc = (int16_t)ni;           // number of A/D converters
    diam = (float)dc;           // cavity diameter (cm)
    temp = (float)tc;           // cavity temperature (C)
    scale = (float)sc;          // response waveform scale factor
    sets = (int16_t)ns;           // number of sweep sets
    adsn = (float)ac;
    da1sn = (float)ds1;
    da2sn = (float)ds2;

    // allocate memory
    s2 = (int16_t *)calloc(npts, sizeof(int16_t));
    s4 = (int32_t *) calloc(npts, sizeof(int32_t));
  
    // convert stimulus waveform to short array
    for (i = 0; i < npts; i++) {
	  s2[i] = (int16_t)round(st[i]);
    }
	
    // set variables
    nw = nc * no * ni;          // number of waveforms
    nv = nw + 18;               // number of variables
   
    sz =strlen(sfn);           // length of stimulus name
	
    vl = sp_var_alloc(nv);

    sp_var_set(vl + 0, ftype,  &nw,   1, 1, "i2");
    sp_var_set(vl + 1, "rate", &rate, 1, 1, "f4");
    sp_var_set(vl + 2, "npts", &npts, 1, 1, "i2");
    sp_var_set(vl + 3, "reps", &reps, 1, 1, "i2");
    sp_var_set(vl + 4, "attn", &attn, 1, 1, "f4");
    sp_var_set(vl + 5, "adsn", &adsn, 1, 1, "f4");
    sp_var_set(vl + 6, "da1sn", &da1sn, 1, 1, "f4");
    sp_var_set(vl + 7, "da2sn", &da2sn, 1, 1, "f4");
    sp_var_set(vl + 8, "mpsn", &mpsn, 1, 1, "f4");
    sp_var_set(vl + 9, "scale", &scale, 1, 1, "f4");
    sp_var_set(vl + 10, "stmfn", sfn, 1, sz, "t");
    sp_var_set(vl + 11, "stim", s2, npts, 1, "i2");
    sp_var_set(vl + 12, "ncav", &ncav, 1, 1, "i2");
    sp_var_set(vl + 13, "nsrc", &nsrc, 1, 1, "i2");
    sp_var_set(vl + 14, "nadc", &nadc, 1, 1, "i2");
    sp_var_set(vl + 15, "diam", &diam, 1, 1, "f4");
    sp_var_set(vl + 16, "temp", &temp, 1, 1, "f4");

    // convert response waveforms to long arrays
    for (j = 0; j < nw; j++) {
        sprintf(wvn, "cal%02d", j + 1);
        for (i = 0; i < npts; i++) {
           s4[i] = (int32_t)round(rs[i + j * npts]);
        }
        sp_var_set(vl + 17 + j, wvn, s4, npts, 1, "i4");
    }
    sp_var_set(vl + nw + 17, "sets", &sets, 1, 1, "i2");
    // save file & clean-up
    sp_mat_save(cfn, vl);   // save variables in MAT file
    sp_var_clear_all();     // deallocate variables
    // free memory
    free(s2);
    free(s4);
}


_FUNC(int) save_cal(int32_t id, float *st, float *rs,
    double fs, double at, double mp, double dc, double tc, double sc,
    int nr, int ns, int np, int nc, int no, int ni, 
    char *sfn,  char *cfn)
{
    char wvn[8];
    double ad_vfs[MAXNCH], da_vfs[MAXNCH];
    float rate, attn, adsn, da1sn, da2sn, mpsn, scale;
    float diam, temp;
    int i, j, sz, nch = 2;
    int32_t *s4;
    int16_t *s2;
    int16_t nv, nw, reps, npts, ncav, nsrc, nadc, sets;
    VAR *vl;
    static char ftype[] = "EMAV_CAL";	

    rate = (float)fs;           // sampling rate
    npts = (int16_t)np;           // number of points
    reps = (int16_t)nr;           // sweeps per set
    attn = (float)at;           // stimulus attenuation
    mpsn = (float)mp;           // microphone sensitivity
    ncav = (int16_t)nc;           // number of cavities
    nsrc = (int16_t)no;           // number of sources
    nadc = (int16_t)ni;           // number of A/D converters
    diam = (float)dc;           // cavity diameter (cm)
    temp = (float)tc;           // cavity temperature (C)
    scale = (float)sc;          // response waveform scale factor
    sets = (int16_t)ns;           // number of sweep sets
    

	// use ARSC soundcard info to set A/D & D/A sensitivities if id != -100
	
	ar_io_open(id, fs, nch, nch);
    ar_get_vfs(id, da_vfs, ad_vfs);
    ar_io_close(id);

    adsn = (float)((pow(2, 23) - 1) / fabs(ad_vfs[0]));
    da1sn = (float)((pow(2, 15) - 1) / fabs(da_vfs[0]));
    da2sn = (float)((pow(2, 15) - 1) / fabs(da_vfs[1]));
    scale *= (float)fabs(ad_vfs[0]);
	
    // allocate memory	
    s2 = (int16_t *) calloc(npts, sizeof(int16_t));
    s4 = (int32_t *) calloc(npts, sizeof(int32_t));
    // convert stimulus waveform to short array
    for (i = 0; i < npts; i++) {
        s2[i] = (int16_t)round(st[i]);
    }
    // set variables
    nw = nc * no * ni;          // number of waveforms
    nv = nw + 18;               // number of variables
    sz = strlen(sfn);           // length of stimulus name
    vl = sp_var_alloc(nv);
    sp_var_set(vl + 0, ftype,  &nw,   1, 1, "i2");
    sp_var_set(vl + 1, "rate", &rate, 1, 1, "f4");
    sp_var_set(vl + 2, "npts", &npts, 1, 1, "i2");
    sp_var_set(vl + 3, "reps", &reps, 1, 1, "i2");
    sp_var_set(vl + 4, "attn", &attn, 1, 1, "f4");
    sp_var_set(vl + 5, "adsn", &adsn, 1, 1, "f4");
    sp_var_set(vl + 6, "da1sn", &da1sn, 1, 1, "f4");
    sp_var_set(vl + 7, "da2sn", &da2sn, 1, 1, "f4");
    sp_var_set(vl + 8, "mpsn", &mpsn, 1, 1, "f4");
    sp_var_set(vl + 9, "scale", &scale, 1, 1, "f4");
    sp_var_set(vl + 10, "stmfn", sfn, 1, sz, "t");
    sp_var_set(vl + 11, "stim", s2, npts, 1, "i2");
    sp_var_set(vl + 12, "ncav", &ncav, 1, 1, "i2");
    sp_var_set(vl + 13, "nsrc", &nsrc, 1, 1, "i2");
    sp_var_set(vl + 14, "nadc", &nadc, 1, 1, "i2");
    sp_var_set(vl + 15, "diam", &diam, 1, 1, "f4");
    sp_var_set(vl + 16, "temp", &temp, 1, 1, "f4");
    // convert response waveforms to long arrays
    for (j = 0; j < nw; j++) {
        sprintf(wvn, "cal%02d", j + 1);
        for (i = 0; i < npts; i++) {
            s4[i] = (int32_t)round(rs[i + j * npts]);
        }
        sp_var_set(vl + 17 + j, wvn, s4, npts, 1, "i4");
    }
    sp_var_set(vl + nw + 17, "sets", &sets, 1, 1, "i2");
    // save file & clean-up
    sp_mat_save(cfn, vl);   // save variables in MAT file
    sp_var_clear_all();     // deallocate variables
    // free memory
    free(s2);
    free(s4);			

	return 0;
}

double _at,_dc,_fs,_df,_mp,_tc;
int _nCav,_npts,_nrep,_styp,_thl,_ths,_wd,_skip,_dt,_reps,_sets;
int _stc, _hpf, _rr;
int _nfreq,_nwave,_nout,_ninn,_id;
float *_fpRN = NULL;
float *_stim = NULL;
/* Usage: Following functions are to be called in following order
	Init - Initializes global variables using values from CU
	Runcal(calId) - for each cavity, call this. calId=0 based number representing each cavity
	SaveCal(numCav) - after Sync-Averaging in each cavity, save results as PRB or CAL file : numcav= # of cavities tested
	DisposeAll - free all resources
*/
static int GetLengthOne(float * st,float *rs, double *lengths)
{//save_cal and calsrc combined. no file output
	char wvn[8];
	double sc;
	double erf[2], ad_vfs[MAXNCH], da_vfs[MAXNCH];
    float rate, attn, adsn, da1sn, da2sn, mpsn, scale;
    float diam, temp;
    int i, j, ns;
    int32_t *s4;
    int16_t *s2;
    int16_t nv, nw, reps, npts, ncav, nsrc, nadc,sets;
    VAR *vl;
	static char ftype[] = "EMAV_CAL";
	char sfn[]="dummy";
	erf[0]=2000;
	erf[1]=8000;
	sc=1/ (2*_reps *(pow(2,23) -1)*_mp);	

    rate = (float)_fs;           // sampling rate
    npts = (int16_t)_npts;           // number of points
    reps = (int16_t)_reps;           // number of reps
    attn = (float)_at;           // stimulus attenuation
    mpsn = (float)_mp;           // microphone sensitivity
    ncav = (int16_t)_nCav;           // number of cavities
    nsrc = (int16_t)_nout;           // number of sources
    nadc = (int16_t)_ninn;           // number of A/D converters
    diam = (float)_dc;           // cavity diameter (cm)
    temp = (float)_tc;           // cavity temperature (C)
    scale = (float)sc;          // response waveform scale factor
	sets = (int16_t)_sets;		//number of sweep sets
    // use ARSC soundcard info to set A/D & D/A sensitivities
	if(ar_io_open(_id,3200,2,2)!=0 ) return 1;
    ar_get_vfs(_id, da_vfs, ad_vfs);
	ar_io_close(_id);
    adsn = (float)((pow(2, 23) - 1) / fabs(ad_vfs[0]));
    da1sn = (float)((pow(2, 15) - 1) / fabs(da_vfs[0]));
    da2sn = (float)((pow(2, 15) - 1) / fabs(da_vfs[1]));
    scale *= (float)fabs(ad_vfs[0]);
    // allocate memory
    s2 = (int16_t *)calloc(npts, sizeof(int16_t));
    s4 = (int32_t *) calloc(npts, sizeof(int32_t));
    // convert stimulus waveform to short array
    for (i = 0; i < npts; i++) {
        s2[i] = (int16_t)round(st[i]);
    }
    // set variables
    nw = ncav * nsrc * nadc;          // number of waveforms
    nv = 18 + nw;               // number of variables
    ns = strlen(sfn);           // length of stimulus name
    vl = sp_var_alloc(nv);
    sp_var_set(vl + 0, ftype,  &nw,   1, 1, "i2");
    sp_var_set(vl + 1, "rate", &rate, 1, 1, "f4");
    sp_var_set(vl + 2, "npts", &npts, 1, 1, "i2");
    sp_var_set(vl + 3, "reps", &reps, 1, 1, "i2");
    sp_var_set(vl + 4, "attn", &attn, 1, 1, "f4");
    sp_var_set(vl + 5, "adsn", &adsn, 1, 1, "f4");
    sp_var_set(vl + 6, "da1sn", &da1sn, 1, 1, "f4");
    sp_var_set(vl + 7, "da2sn", &da2sn, 1, 1, "f4");
    sp_var_set(vl + 8, "mpsn", &mpsn, 1, 1, "f4");
    sp_var_set(vl + 9, "scale", &scale, 1, 1, "f4");
    sp_var_set(vl + 10, "stmfn", sfn, 1, ns, "t");
    sp_var_set(vl + 11, "stim", s2, npts, 1, "i2");
    sp_var_set(vl + 12, "ncav", &ncav, 1, 1, "i2");
    sp_var_set(vl + 13, "nsrc", &nsrc, 1, 1, "i2");
    sp_var_set(vl + 14, "nadc", &nadc, 1, 1, "i2");
    sp_var_set(vl + 15, "diam", &diam, 1, 1, "f4");
    sp_var_set(vl + 16, "temp", &temp, 1, 1, "f4");
    // convert response waveforms to long arrays
    for (j = 0; j < nw; j++) {
        sprintf(wvn, "cal%02d", j + 1);
        for (i = 0; i < npts; i++) {
            s4[i] = (int32_t)round(rs[i + j * npts]);
        }
        sp_var_set(vl + 17 + j, wvn, s4, npts, 1, "i4");
    }
    sp_var_set(vl+nw+17,"sets",&sets,1,1,"i2");

    GetLength1(vl,erf,lengths);   
   // save file & clean-up
    sp_var_clear_all();     // deallocate variables
    // free memory
    free(s2);
    free(s4);
    return 0;	

}

_FUNC (int) Init(CAV_USAGE *cu, int soundid)
{
    _at = cu->at;            // attenuation
    _npts = cu->npts;        // buffer size
    _nCav = cu->nCav;        // number of cavities
    _dc = 0.8;//cu->dc ;     // cavity diameter
    _thl = cu->thl;          // thl flag
    _styp = cu->styp;        // stimulus type 0=noise, 1=chirp
    _fs = cu->fs;            // sampling rate
    _nrep = cu->nrep;        // number of sweeps
    _sets = 2;               // number of sweep sets
    _reps = _nrep/(2*_sets); // sweeps per set
    _wd = cu->wd;            // write data
    _tc = 25;                // cavity temperature
    _ths = cu->ths;          // ths flag
    _id = soundid;           // soundcard 
    // microphone sensitivity (v/pa)
    if ((cu->mp>0.001)&&(cu->mp<1000)) {
        _mp = cu->mp;
    } else {
	_mp = 0.5;
    }
    _nout = 2;
    _ninn = 1; 
    _skip=1;
    _dt = 1; //detrend
    _nfreq = _npts/2+1;      // number of frequencies
    _nwave = _nCav * _nout * _ninn;  // number of waveforms
    _df = (_fs/_npts)/1000;
    free_null(_stim);
    free_null(_fpRN);
    _stim = (float *) calloc(_npts ,sizeof(float));
    _fpRN = (float *) calloc(_npts * _nwave,sizeof(float));
    _stc=0; 
    _hpf=0; 
    _rr=1;
	
    gen_stim(_stim,_npts,_styp); //create stim
	
    return (0);
}
_FUNC (int) InitABR(int SoundId)
{
    _at = 20;			// attenuation
    _npts = 8192;		// buffer size
    _nCav = 5;        // number of cavities
    _dc = 0.8;//cu->dc ;           // cavity diameter
    _thl = 0;		    // thl flag
    _styp = 1;	    // stimulus type 0=noise, 1=chirp
    _fs = 48000;            // sampling rate
    // _nrep = cu->nrep;        // number of sweeps
    _reps = 4;				//sweeps per set
    _sets = 2;				// number of sweep sets
    _nrep = _reps*_sets*2;  //number of sweeps
    _wd = 0;		    // write data
    _tc = 25;				//cavity temperature
    _ths = 0;			//ths flag
    _id = SoundId;			//soundcard 
    _mp =0.5;				// microphone sensitivity (v/pa)
    _nout = 2;
    _ninn = 2; 
    _skip=1;
    _dt = 1; //detrend
    _nfreq = _npts/2+1;  //number of frequencies
    _nwave = _nCav * _nout * _ninn;  // number of waveforms
    _df = (_fs/_npts)/1000;
    free_null(_stim);
    free_null(_fpRN);
    _stim = (float *)calloc(_npts ,sizeof(float));
    _fpRN = (float * ) calloc(_npts * _nwave,sizeof(float));
	
    gen_stim(_stim,_npts,_styp); //create stim
	
    return (0);

}

_FUNC (int)RunCal(int calId)
{
    int j,k;
    float *rs,*H;
    char fn[MAXLEN];
    H = (float *)calloc(_nfreq*2,sizeof(float));
    for(j=0;j<_nout;j++) {
        for(k=0;k<_ninn;k++) {
            rs = _fpRN +((calId *_nout +j)*_ninn +k)*_npts;
            get_response(_npts,_nrep,j,k,_skip,_fs,_at,_id,_stim,rs);
            if(_dt) detrend(rs,_npts);
            sp_transfer(_stim,rs,_npts,H);
            sprintf(fn, "cav%d%d%d.txt", calId + 1, j + 1, k + 1);
            if (_wd) freq_resp(H, _nfreq, _df, fn);         
        }//end k
    }//end j
    free(H);
    return (0);
}

_FUNC (int)GetResponse(float *resp)
{
    int j,k;
    float *rs;
    for(j=0;j<_nout;j++) {
        for(k=0;k<_ninn;k++) {
            rs = _fpRN +((_nout +j)*_ninn +k)*_npts;
            get_response(_npts,_nrep,j,k,_skip,_fs,_at,_id,_stim,rs);
        }//end k
    }//end j
    memcpy(resp, _fpRN, _nout * _ninn * _npts * sizeof(float));
    return (0);
}

_FUNC (int)SaveCal(int numCav, double *erf)
{
    char sfn[MAXLEN],cfn[MAXLEN];
    double sc;		

    strcpy(sfn,_styp? "~lsw":"~bbn");
    if(numCav >1) {
        genfn("C","PRB",cfn);
    } else {
        genfn("C","CAL",cfn);				
    }
    sc = 1/(2*_reps*(pow(2,23)-1)*_mp);
    save_cal(_id,_stim,_fpRN,_fs,_at,_mp,_dc,_tc,sc,
        _reps,_sets,_npts,numCav,_nout,_ninn,sfn,cfn);
    calsrc(cfn, erf, _thl, _wd, _stc, _hpf, _rr);

    return 0;
}

_FUNC (int)GetLengths(int numCav,double *lengths)
{
    char sfn[MAXLEN],cfn[MAXLEN];
    double sc; 
    double erf[2];

    erf[0]=2000;
    erf[1]=8000;
    strcpy(sfn,_styp? "~lsw":"~bbn");
    genfn("C","PRB",cfn);
	
    sc= 1/(2*_reps *(pow(2,23) -1)*_mp);
    save_cal(_id,_stim,_fpRN,_fs,_at,_mp,_dc,_tc,sc,
        _reps,_sets,_npts,numCav,_nout,_ninn,sfn,cfn);
    calsrc(cfn, erf, _thl, _wd, _stc, _hpf, _rr);
	
    return 0;
}

_FUNC(int)GetSingleLength(double *lengths)
{
    double erf[2];

    erf[0]=2000;
    erf[1]=8000;

    if (GetLengthOne(_stim,_fpRN,lengths)!=0 ) return 1;
	
    return 0;
}

_FUNC (int) DisposeAll(void)
{
    free_null(_stim);
    free_null(_fpRN);
    return (0);
}

_FUNC (int) myInt(int n)
{
    return n+3;
}

_FUNC (void) run(void)
{
    int ac = 0;
    //char **av;
    //main(0,av);
    return;
}

_FUNC(int)NumberOfDevices()
{
    return ar_num_devs();
}

_FUNC(int)DeviceName(int dev,char *name,int len)
{
    return ar_dev_name(dev,name,len);
}

#else // MAKEDLL

static void
get_vfs(int32_t id, double fs, double *da_vfs, double *ad_vfs)
{
    int nch = 2;

    // get A/D & D/A sensitivities from ARSC soundcard info
    ar_io_open(id, fs, nch, nch);
    ar_get_vfs(id, da_vfs, ad_vfs);
    ar_io_close(id);
}

static void
save_cal(double *da_vfs, double *ad_vfs, float *st, float *rs,
    double fs, double at, double mp, double dc, double tc, double sc,
    int nr, int ns, int np, int nc, int no, int ni, 
    char *sfn, char *cfn)
{
    char wvn[20];
    float rate, attn, adsn, da1sn, da2sn, mpsn, scale;
    float diam, temp;
    int i, j, sz;
    int32_t *s4;
    int16_t *s2;
    int16_t nv, nw, reps, npts, ncav, nsrc, nadc, sets;
    VAR *vl;
    static char ftype[] = "EMAV_CAL";

    rate = (float)fs;           // sampling rate
    npts = (int16_t)np;           // number of points
    reps = (int16_t)nr;           // sweeps per set
    attn = (float)at;           // stimulus attenuation
    mpsn = (float)mp;           // microphone sensitivity
    ncav = (int16_t)nc;           // number of cavities
    nsrc = (int16_t)no;           // number of sources
    nadc = (int16_t)ni;           // number of A/D converters
    diam = (float)dc;           // cavity diameter (cm)
    temp = (float)tc;           // cavity temperature (C)
    scale = (float)sc;          // response waveform scale factor
    sets = (int16_t)ns;           // number of sweep sets
    adsn = (float)((pow(2, 23) - 1) / fabs(ad_vfs[0]));
    da1sn = (float)((pow(2, 15) - 1) / fabs(da_vfs[0]));
    da2sn = (float)((pow(2, 15) - 1) / fabs(da_vfs[1]));
    // convert stimulus waveform to short array
    s2 = (int16_t *) calloc(npts, sizeof(int16_t));
    for (i = 0; i < npts; i++) {
        s2[i] = (int16_t)round(st[i]);
    }
    // set variables
    nw = nc * no * ni;          // number of waveforms
    nv = nw + 18;               // number of variables
    sz = strlen(sfn);           // length of stimulus name
    vl = sp_var_alloc(nv);
    sp_var_set(vl + 0, ftype,  &nw,   1, 1, "i2");
    sp_var_set(vl + 1, "rate", &rate, 1, 1, "f4");
    sp_var_set(vl + 2, "npts", &npts, 1, 1, "i2");
    sp_var_set(vl + 3, "reps", &reps, 1, 1, "i2");
    sp_var_set(vl + 4, "attn", &attn, 1, 1, "f4");
    sp_var_set(vl + 5, "adsn", &adsn, 1, 1, "f4");
    sp_var_set(vl + 6, "da1sn", &da1sn, 1, 1, "f4");
    sp_var_set(vl + 7, "da2sn", &da2sn, 1, 1, "f4");
    sp_var_set(vl + 8, "mpsn", &mpsn, 1, 1, "f4");
    sp_var_set(vl + 9, "scale", &scale, 1, 1, "f4");
    sp_var_set(vl + 10, "stmfn", sfn, 1, sz, "t");
    sp_var_set(vl + 11, "stim", s2, npts, 1, "i2");
    sp_var_set(vl + 12, "ncav", &ncav, 1, 1, "i2");
    sp_var_set(vl + 13, "nsrc", &nsrc, 1, 1, "i2");
    sp_var_set(vl + 14, "nadc", &nadc, 1, 1, "i2");
    sp_var_set(vl + 15, "diam", &diam, 1, 1, "f4");
    sp_var_set(vl + 16, "temp", &temp, 1, 1, "f4");
    // convert response waveforms to long arrays
    s4 = (int32_t *) calloc(npts, sizeof(int32_t));
    for (j = 0; j < nw; j++) {
        sprintf(wvn, "cal%02d", j + 1);
        for (i = 0; i < npts; i++) {
            s4[i] = (int32_t)round(rs[i + j * npts]);
        }
        sp_var_set(vl + 17 + j, wvn, s4, npts, 1, "i4");
    }
    sp_var_set(vl + nw + 17, "sets", &sets, 1, 1, "i2");
    // save file & clean-up
    sp_mat_save(cfn, vl);   // save variables in MAT file
    sp_var_clear_all();     // deallocate variables
    free(s2);
    free(s4);
}

static void
done(int pbe)
{
    if (pbe) {
        printf("Press Enter...\n");
	getchar();
    }
    exit(0);
}


static void
usage()
{
    printf("\n");
    printf("usage: cavcal [option|PRBfile]\n");
    printf("\n");
    printf("options:\n");
    printf(" -a N  - attenuation (dB) = N [20]\n");
    printf(" -b N  - buffer size = N [8192]\n");
    printf(" -c N  - number of cavities = N [5]\n");
    printf(" -d N  - cavity diameter (cm) = N [0.8]\n");
    printf(" -f1 N - min. error frequency (Hz) = N [500]\n");
    printf(" -f2 N - max. error frequency (Hz) = N [8000]\n");
    printf(" -h  N - apply high-pass filter at N (Hz) [0]\n");
    printf(" -i    - device_identifier = N [0]\n");
    printf(" -l    - create THL file for each cavity\n");
    printf(" -m N  - microphone sensistivity (V/Pa) = N [0.5]\n");
    printf(" -n    - use white-noise stimulus\n");
    printf(" -p    - pause before exit\n");
    printf(" -r N  - sampling_rate = N [32000]\n");
    printf(" -R    - disable ideal-reflection reduction\n");
    printf(" -s N  - sweeps_per_set = N [4]\n");
    printf(" -t N  - cavity temperature (C) = N [25]\n");
    printf(" -w    - create TXT files for plotting data\n");
    printf(" -z N  - sets = N [2]\n");
    printf("\n");
    printf("nrep = sweeps_per_set * sets * 2\n");
    exit(0);
}

int
main(int ac, char **av)
{
    char dn[MAXLEN], fn[MAXLEN], cfn[MAXLEN], sfn[MAXLEN];
    char s1[MAXLEN], s2[MAXLEN], s3[MAXLEN];
    double ad_vfs[MAXNCH], da_vfs[MAXNCH];
    double df, sc, fs, at, adsn, mpsn, dc, tc, erf[2];
    float *st, *rs, *rn, *H;
    int i, j, k, er, c, nf, nw, npts, reps, sets, nrep;
    static int stc = 0;         // set temperature for calsrc
    static int ncav = 5;        // number of cavities
    static int nout = 2;        // number of channels out
    static int ninn = 1;        // number of channels in
    static int skip = 1;        // number of sweeps skipped
    static int styp = 1;        // simulus type: 0=noise, 1=chirp
    static int dt = 1;          // detrend response ?
    static int wd = 0;          // write data ?
    static int rr = 1;          // ideal-reflection reduction ?
    static int hpf = 0;         // apply high-pass filter ?
    static int thl = 0;         // create THL ?
    static int pbe = 0;         // pause before exit ?
    static int32_t id = 0;      // device identifier

    // [EMAV equivalent parameters]
    mpsn = 0.5;     // microphone sensitivity (V/Pa)    [SYSTEM:MP_sensitivity]
    fs = 32000;     // sampling rate (Hz)               [DPOAE:rate]
    npts = 8192;    // buffer size                      [DPOAE:size]
    reps = 4;       // sweeps per set                   [DPOAE:sweeps]
    sets = 2;       // number of sweep sets             [DPOAE:sets]
    at = 20;        // atten. from max. amplitude (dB)  [DPOAE:atten]
    dc = 0.8;       // cavity diameter (cm)             [PROBE:diacav]
    tc = 25;        // cavity temperature (C)           [PROBE:temp]
    erf[0] = 2000;  // min. error frequency (Hz)        [PROBE:f1_erf]
    erf[1] = 8000;  // max. error frequency (Hz)        [PROBE:f2_erf]

    cavcal_version(s1, s2, s3);
    printf("%s\n", s1);
    printf("%s\n", s2);
    printf("%s\n", s3);
    printf("-------------------\n");

    // parse command-line
    while (ac > 1) {
        if (_access(av[1],0) == 0) {                // does file exist ?
            calsrc(av[1], erf, thl, wd, stc, hpf, rr);  // create THS file
            if (ac < 3) {
		done(pbe);
            }
        } else if (av[1][0] == '-') {               // options ?
	    if (av[1][1] == 'a') {
                at = atof(av[2]);                   // attenuation
	    } else if (av[1][1] == 'b') {
                npts = atoi(av[2]);                 // buffer size
	    } else if (av[1][1] == 'c') {
                ncav = atoi(av[2]);                 // number of cavities
	    } else if (av[1][1] == 'd') {
                dc = atof(av[2]);                   // cavity diameter
	    } else if (av[1][1] == 'f' && av[1][2] == '1') {
                erf[0] = atof(av[2]);               // min. error freq.
	    } else if (av[1][1] == 'f' && av[1][2] == '2') {
                erf[1] = atof(av[2]);               // max. error freq.
	    } else if (av[1][1] == 'h') {
                if (ac > 2) {
                    hpf = atoi(av[2]);                  // high-pass filter
                    printf("high-pass_filter = %d (Hz)\n", hpf);
                } else {
                    usage();
                }
	    } else if (av[1][1] == 'i') {
                id = atoi(av[2]);                   // device identifier
	    } else if (av[1][1] == 'l') {
                thl = 1;		            // thl flag
	    } else if (av[1][1] == 'm') {
                mpsn = atof(av[2]);                 // microphone sensitivity
	    } else if (av[1][1] == 'n') {
                styp = 0;		            // white-noise stimulus
	    } else if (av[1][1] == 'p') {
                pbe = 1;		            // pause before exit
	    } else if (av[1][1] == 'r') {
                fs = atof(av[2]);                   // sampling_rate
	    } else if (av[1][1] == 'R') {
                rr = 0;                             // disable ideal-reflection reduction
	    } else if (av[1][1] == 's') {
                reps = atoi(av[2]);                 // sweeps_per-set
	    } else if (av[1][1] == 't') {
                tc = atof(av[2]);                   // cavity temperature
                stc = round(tc);
	    } else if (av[1][1] == 'w') {
                wd = 1;			            // write data
	    } else if (av[1][1] == 'z') {
                sets = atoi(av[2]);                 // sets
	    } else {
		usage();
	    }
	    // bump options with arguments
	    if (strchr("abcdfhimrstz",av[1][1])) {
		ac--;
		av++;
	    }
        } else {
            usage();
        } 
        ac--;
        av++;
    }

    // show parameter values
    nrep = reps * sets * 2;
    printf("ncav=%d npts=%d nrep=%d*%d*2\n", ncav, npts, reps, sets);
    printf("rate=%.0f diam=%0.2f temp=%.0f ", fs, dc, tc);
    printf("atten=%.0f mp=%.3f\n\n", at, mpsn);
    if (fs < 9000) {
	fprintf(stderr, ">>> sampling rate is too low.\n");
	exit(1);
    }

    nf = npts / 2 + 1;                         // number of frequencies
    nw = ncav * nout * ninn;                   // number of waveforms
    if (id == 0) {
        id = ar_find_dev(ARSC_PREF_SYNC);      // find preferred i/o device
    }
    er = ar_dev_name(id, dn, MAXLEN);
    if (er) {
        printf("   No i/o device.\n");
        return (1);
    }
    printf("   i/o device = %s\n\n", dn);

    df = (fs / npts) / 1000;
    H = (float *)calloc(nf * 2, sizeof(float));
    st = (float *)calloc(npts, sizeof(float));
    rn = (float *)calloc(npts * nw, sizeof(float));
    gen_stim(st, npts, styp);
    for (i = 0; i < ncav; i++) {
        printf("Press Enter to begin cavity %d: ", i + 1);
        c = getchar();
        if (c == 'x') break;
        for (j = 0; j < nout; j++) {
            for (k = 0; k < ninn; k++) {
                rs = rn + ((i * nout + j) * ninn + k) * npts;
                get_response(npts, nrep, j, k, skip, fs, at, id, st, rs);
		if (dt) detrend(rs, npts);
                sp_transfer(st, rs, npts, H);   // transfer function
                sprintf(fn, "cav%d%d%d.txt", i + 1, j + 1, k + 1);
                if (wd) freq_resp(H, nf, df, fn);
                msleep(500);                    // wait 500 ms
            }
        }
    }
    strcpy(sfn, styp ? "~lsw" : "~bbn");        // stimulus name
    if (ncav > 1) {
	genfn("C","PRB",cfn);
    } else {
	genfn("C","CAL",cfn);
    }
    get_vfs(id, fs, da_vfs, ad_vfs);
    adsn = (pow(2, 23) - 1) / fabs(ad_vfs[0]);
    // EMAV: scale = 1 / (2 * reps * adsn * mpsn)
    sc = 1 / (2 * reps * adsn * mpsn);
    save_cal(da_vfs, ad_vfs, st, rn, fs, at, mpsn, dc, tc, sc,
        reps, sets, npts, ncav, nout, ninn, sfn, cfn);
    printf("Wrote %s.\n", cfn);
    free(H);
    free(st);
    free(rn); 
    if (ncav > 2) {
        calsrc(cfn, erf, thl, wd, 0, hpf, rr);      // create THS file
    }

    done(pbe);
    return (0);
}

#endif //MAKEDLL
