// mptrans.c - microphone-sensitivity transfer function

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sigpro.h>
#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#define _access access
#define _hypot  hypot
#endif

#define FNSZ          256
#define MAXLINE       256
#define MAXPOSINT     32767
#define free_null(x)  if(x){free(x);x=NULL;}

static char    mps_ifn[FNSZ];
static double  mps_in = 0;
static double  mps_df = 0;
static float  *mps_tf = NULL;
static int     mps_nf = 0;

/***********************************************************************/

#ifdef _DEBUG
static void
mp_transfer_write1(char *ofn, float *fr, float *mg, float *ph, int nf)
{
    int i;
    FILE *fp;

    if (nf > 0) {
        fp = fopen(ofn, "wt");
	fprintf(fp, "; %s\n", ofn);
	fprintf(fp, "; frequency_range = %.0f - %.0f Hz\n", fr[0], fr[nf - 1]);
	fprintf(fp, "; MP_transfer = %s\n", mps_ifn);
	fprintf(fp, "; MP_sensitivity = %.3f\n", mps_in);
	fprintf(fp, ";    f     mg     ph\n");
	for (i = 0; i < nf; i++) {
	    fprintf(fp, " %6.0f %7.4f %6.3f\n", fr[i], mg[i], ph[i]);
	}
	fclose(fp);
    }
}

static void
mp_transfer_write2(char *ofn)
{
    double fr, mg, ph, pp, tpi;
    int i, ir, ii;
    FILE *fp;

    if (mps_nf > 0) {
	fr = mps_df * (mps_nf - 1);
        fp = fopen(ofn, "wt");
	fprintf(fp, "; %s\n", ofn);
	fprintf(fp, "; frequency_range = 0 - %.0f Hz\n", fr);
	fprintf(fp, "; MP_transfer = %s\n", mps_ifn);
	fprintf(fp, "; MP_sensitivity = %.3f\n", mps_in);
	fprintf(fp, ";    f     mg     ph\n");
	tpi = 2 * M_PI;
	pp = 0;
	for (i = 0; i < mps_nf; i++) {
	    ir = i * 2;
	    ii = i * 2 + 1;
	    fr = i * mps_df;
	    mg = 1 / _hypot(mps_tf[ir], mps_tf[ii]);
	    ph = -atan2(mps_tf[ii], mps_tf[ir]) / tpi;
	    while ((ph - pp) > 0.5) ph -= 1;
	    fprintf(fp, " %6.0f %7.4f %6.3f\n", fr, mg, ph);
	    pp = ph;
	}
	fclose(fp);
    }
}
#endif // _DEBUG

static void
mp_transfer_interpolate(float *fr, float *mg, float *ph, int nf)
{
    double a, hfr, hmg, hph, tpi;
    int i, ir, ii, j;

    if ((nf > 0) && (mps_nf > 0)) {
	free_null(mps_tf);
	mps_tf = (float *) calloc(mps_nf * 2, sizeof(float));
	tpi = 2 * M_PI;
	j = 0;
	for (i = 0; i < mps_nf; i++) {
	    ir = i * 2;
	    ii = i * 2 + 1;
	    hfr = mps_df * i;
	    while ((j < (nf - 1)) && (fr[j + 1] <= hfr)) j++;
	    if ((hfr <= fr[j]) || (j == (nf - 1))) {
                hmg = mg[j];
                hph = ph[j];
	    } else {
		a = (hfr - fr[j]) / (fr[j + 1] - fr[j]);
                hmg = mg[j] * (1 - a) + mg[j + 1] * a;
                hph = ph[j] * (1 - a) + ph[j + 1] * a;
	    }
	    mps_tf[ir] = (float) (cos(-tpi * hph) / hmg);
	    mps_tf[ii] = (float) (sin(-tpi * hph) / hmg);
	}
    }
}

// read microphone-sensitivity transfer function from MAT file
static int
mp_transfer_mat(char *fn, double mpsn)
{
    double sens_mp, fmax, df, mr, pp, tpi;
    float *resp, *fr, *mg, *ph;
    int i, ir, ii, m, n, nf;
    VAR *v;
    static float mgmn = 0.001F;
    static float mgmx = 1000.F;

    v = sp_mat_load(fn);
    if (!v) {
	goto err_return;
    }
    if (sp_var_find(v, "sens_mp") < 0) {
	sens_mp = mpsn;
    } else {
        sens_mp = sp_var_f8(v, "sens_mp");
    }
    if (sens_mp == 0) {
	sens_mp = mpsn;
    }
    if (sp_var_find(v, "fmax") < 0) {
	goto err_return;
    } else {
        fmax = sp_var_f8(v, "fmax");
    }
    if (sp_var_find(v, "resp") < 0) {
	goto err_return;
    } else {
        i = sp_var_find(v, "resp");
        n = v[i].rows * v[i].cols;
	m = n + 2;
	resp = (float *) calloc(m, sizeof(float));
        memcpy(resp, v[i].data, m * sizeof(float));
        sp_rcfft(resp, n);
	tpi = 2 * M_PI;
	mr = sens_mp / mpsn;
	df = 2 * fmax / n;
	nf = m / 2;
        fr = (float *) calloc(nf * 3, sizeof(float));
        mg = fr + nf;
        ph = mg + nf;
	pp = 0;
	for (i = 0; i < nf; i++) {
            ir = i * 2;
            ii = i * 2 + 1;
            fr[i] = (float) (i * df);
            mg[i] = (float) (_hypot(resp[ir], resp[ii]) * mr);
            ph[i] = (float) (atan2(resp[ii], resp[ir]) / tpi);
	    if (mg[i] < mgmn) mg[i] = mgmn;
	    if (mg[i] > mgmx) mg[i] = mgmx;
	    while((ph[i] - pp) > 0.5) ph[i] -= 1;
	    pp = ph[i];
        }
        mp_transfer_interpolate(fr, mg, ph, nf);
        strncpy(mps_ifn, fn, FNSZ);
        mps_in = sens_mp;
#ifdef _DEBUG
        mp_transfer_write1("mp_transfer1.txt", fr, mg, ph, nf);
        mp_transfer_write2("mp_transfer2.txt");
#endif
        free(resp);
        free(fr);
        sp_var_clear(v);
        return (1);
    }
err_return:
    sp_var_clear(v);
    return (0);
}

// read microphone-sensitivity transfer function from TXT file
static int
mp_transfer_txt(char *fn, double mpsn)
{
    char line[MAXLINE];
    float *fr, *mg, *ph;
    int  nf = 0;
    FILE *fp;

    if (_access(fn, 0) == 0) {	// read mp_sens from file
	fp = fopen(fn, "rt");
	while (fgets(line, MAXLINE, fp) != NULL && nf < MAXPOSINT) {
	    if (line[0] > 0 && line[0] != ';') {
		nf++;
    	    }
	}
	fclose(fp);
	if (nf > 0) {
	    fr = (float *) calloc(nf * 3, sizeof(float));
	    mg = fr + nf;
	    ph = mg + nf;
	    nf = 0;
	    fp = fopen(fn, "rt");
	    while (fgets(line, MAXLINE, fp) != NULL && nf < MAXPOSINT) {
		if (line[0] > 0 && line[0] != ';') {
		    fr[nf] = mg[nf] = ph[nf] = 0;
		    // assume: fr=Hz, ph=cyc
		    sscanf(line, " %f %f %f", &fr[nf], &mg[nf], &ph[nf]);
		    nf++;
    		}
	    }
	    fclose(fp);
            mp_transfer_interpolate(fr, mg, ph, nf);
            mp_transfer_interpolate(fr, mg, ph, nf);
            strncpy(mps_ifn, fn, FNSZ);
            mps_in = 1;
#ifdef _DEBUG
            mp_transfer_write1("mp_transfer1.txt", fr, mg, ph, nf);
            mp_transfer_write2("mp_transfer2.txt");
#endif
	    free(fr);
	}
    }
    return (nf);
}

// initialize microphone-sensitivity transfer function
void
microphone_transfer_init(double df, double mpsn, int nf, int msg)
{
    char *fn;

    if ((nf == mps_nf) && (df == mps_df)) {
	return;
    }
    fn = getenv("MP_transfer");
    if (fn == NULL) {
        return;
    }
    mps_nf = nf;
    mps_df = df;
    if (mp_transfer_mat(fn, mpsn)) {
        if (msg) printf("MP_transfer: %s\n", fn);
        return;
    }
    if (mp_transfer_txt(fn, mpsn)) {
        if (msg) printf("MP_transfer: %s\n", fn);
        return;
    }
    free_null(mps_tf);
    mps_nf = 0;
    mps_df = 0;
    return;
}

// compensate for frequency-dependence of microphone sensitivity
void
microphone_transfer(float *p)
{
    float ftemp, *H;
    int i, ir, ii;

    if (mps_tf) {
        H = mps_tf;
        for (i = 0; i < mps_nf; i++) {
            ir = i * 2;
	    ii = i * 2 + 1;
	    ftemp = H[ir] * p[ir] - H[ii] * p[ii];
	    p[ii] = H[ir] * p[ii] + H[ii] * p[ir];
	    p[ir] = ftemp;
        }
    }
}


